# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>


from __future__ import annotations
from typing import Any, Iterable, TypeVar
import logging
from pycyphal.presentation import Subscriber
from pycyphal.presentation.subscription_synchronizer.monotonic_clustering import MonotonicClusteringSynchronizer
from ._sync import SynchronizerOutput, Synchronizer

T = TypeVar("T")


def make_sync_monoclust(
    subscribers: Iterable[Subscriber[Any]],
    *,
    f_key: MonotonicClusteringSynchronizer.KeyFunction,
    tolerance_minmax: tuple[float, float],
) -> Synchronizer:
    tolerance_minmax = float(tolerance_minmax[0]), float(tolerance_minmax[1])
    sync = MonotonicClusteringSynchronizer(subscribers, f_key=f_key, tolerance=max(tolerance_minmax))
    prev_key: Any = None

    async def fun(output: SynchronizerOutput) -> None:
        nonlocal prev_key
        # noinspection PyTypeChecker
        async for synchronized_group in sync:
            key = sum(f_key(x[0]) for x in synchronized_group) / len(synchronized_group)
            output(synchronized_group)
            if prev_key is not None:
                sync.tolerance = _clamp(
                    tolerance_minmax,
                    (sync.tolerance + _tolerance_from_key_delta(prev_key, key)) * 0.5,
                )
            _logger.info("Tolerance autotune: %r", sync.tolerance)
            prev_key = key

    return fun


def _tolerance_from_key_delta(old: T, new: T) -> T:
    return (new - old) * 0.5  # type: ignore


def _clamp(lo_hi: tuple[T, T], val: T) -> T:
    lo, hi = lo_hi
    return min(max(lo, val), hi)  # type: ignore


_logger = logging.getLogger(__name__)



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
from typing import Any, Dict, Callable, TextIO
import time
import ruamel.yaml
import ruamel.yaml.constructor
import yakut
from ._loader import Loader


class EmbeddedExpressionError(ValueError):
    """
    An invalid embedded expression found in the YAML document.
    """


class EvaluableLoader(Loader):
    """
    This is like regular loader except that it feeds scalars tagged as ``!$`` into :func:`eval`
    and substitutes them with the evaluation result.
    """

    EXPRESSION_TAG = "!$"
    """
    Evaluable expressions are tagged with this YAML tag.
    """

    def __init__(self, evaluation_context: Dict[str, Any]) -> None:
        """
        :param evaluation_context: Objects that will be available to the evaluated expressions as their global scope.
        """
        super().__init__()
        self._evaluation_context = evaluation_context.copy()

        class ConstructorWrapper(ruamel.yaml.constructor.RoundTripConstructor):  # type: ignore
            """
            New class to avoid global state: https://stackoverflow.com/questions/67041211
            """

        self._impl.Constructor = ConstructorWrapper
        self._impl.constructor.add_constructor(self.EXPRESSION_TAG, construct_embedded_expression)
        self._impl.constructor.add_multi_constructor("", catch_unknown_tags)

    @property
    def evaluation_context(self) -> Dict[str, Any]:
        """
        This property allows the user to modify the evaluation context after the instance is constructed.
        """
        return self._evaluation_context

    def load(self, text: str | TextIO, **evaluation_context: Any) -> Any:
        """
        Loads and evaluates the evaluable YAML in one operation.
        It is not recommended to use this method if the same YAML document needs to be evaluated multiple times
        (perhaps with different context values); for that, see :meth:`load_unevaluated`.
        This is a mere shortcut for ``self.load_unevaluated(doc)(**evaluation_context)``.
        """
        return self.load_unevaluated(text)(**evaluation_context)

    def load_unevaluated(self, text: str | TextIO) -> Callable[..., Any]:
        """
        Loads the document without evaluation.
        The result is a closure that accepts keyword arguments that extend/override the evaluation context
        passed to the constructor.
        The result of that closure is the evaluated document.
        This way allows you to evaluate the same document with different arguments without re-parsing it from scratch.
        """
        root = self._impl.load(text)

        def evaluate(**kw: Any) -> Any:
            ctx = self._evaluation_context.copy()
            ctx.update(kw)

            def traverse(obj: Any) -> Any:
                if isinstance(obj, dict):
                    return {key: traverse(value) for key, value in obj.items()}
                if isinstance(obj, (list, tuple, set)):
                    return list(map(traverse, obj))
                if isinstance(obj, (bool, int, float, str, bytes)) or obj is None:
                    return obj
                if isinstance(obj, EmbeddedExpression):
                    return obj.evaluate(ctx)
                raise TypeError(f"Unexpected object type: {type(obj).__name__}")  # pragma: no cover

            return traverse(root)

        return evaluate


class EmbeddedExpression:
    """
    An evaluable expression embedded into a YAML document.
    """

    def __init__(self, source_text: str) -> None:
        self._source_text = source_text
        self._code = compile(self._source_text, "<embedded-yaml-expression>", "eval")

    def evaluate(self, evaluation_context: Dict[str, Any]) -> Any:
        started_at = time.monotonic()
        result = eval(self._code, evaluation_context)
        elapsed = time.monotonic() - started_at
        _logger.debug("Expression evaluated successfully in %.3f sec: %s --> %r", elapsed, self, result)
        return result

    def __repr__(self) -> str:
        return repr(self._source_text)


def construct_embedded_expression(_constructor: ruamel.yaml.Constructor, node: ruamel.yaml.Node) -> EmbeddedExpression:
    _logger.debug("Loading embedded expression from node %r", node)
    if not isinstance(node, ruamel.yaml.ScalarNode) or not isinstance(node.value, str):
        raise EmbeddedExpressionError("Embedded expression must be a YAML string")
    try:
        out = EmbeddedExpression(node.value)
    except Exception as ex:
        raise EmbeddedExpressionError(f"Could not load embedded expression from node {node}: {ex}") from ex
    _logger.debug("Successfully constructed embedded expression: %s", out)
    return out


def catch_unknown_tags(_constructor: ruamel.yaml.Constructor, tag: str, node: ruamel.yaml.Node) -> None:
    raise ValueError(f"Unsupported YAML tag {tag!r} encountered in node {node!r}")


_logger = yakut.get_logger(__name__)


def _unittest_eval() -> None:
    import pytest

    loader = EvaluableLoader({"one": 1, "two": 2})
    out = loader.load(
        "{a: 456, b: !$ one + 5, c: [!$ two, !$ foo - two]}",
        foo=3,
    )
    print(out)
    assert out == {"a": 456, "b": 6, "c": [2, 1]}

    evaluator = loader.load_unevaluated("!$ '[one, two, three]'")
    assert evaluator(three=3) == [1, 2, 3]
    loader.evaluation_context["one"] = 11
    loader.evaluation_context["two"] = 222
    assert evaluator(three=-3) == [11, 222, -3]

    with pytest.raises(EmbeddedExpressionError, match=r"(?i).*YAML string.*"):
        loader.load("baz: !$ []")

    with pytest.raises(EmbeddedExpressionError, match=r"(?i).*expression.*"):
        loader.load("baz: !$ 0syntax error")

    with pytest.raises(ValueError, match=r"(?i).*YAML tag.*"):
        loader.load("baz: !bad 123")



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from ._app_descriptor import AppDescriptor as AppDescriptor
from ._cmd import file_server as file_server



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import dataclasses
from typing import TYPE_CHECKING, Union, Callable
import pycyphal
import yakut
from ._directive import Directive, RegisterDirective

if TYPE_CHECKING:
    import pycyphal.application
    from uavcan.register import Access_1


@dataclasses.dataclass(frozen=True)
class Tag:
    def __eq__(self, other: object) -> bool:
        """
        >>> TypeCoercionFailure("foo") == TypeCoercionFailure("bar")
        True
        >>> TypeCoercionFailure("") == Timeout()
        False
        """
        return issubclass(type(self), type(other)) and issubclass(type(other), type(self))


@dataclasses.dataclass(frozen=True)
class TypeCoercionFailure(Tag):
    msg: str

    __eq__ = Tag.__eq__


@dataclasses.dataclass(frozen=True)
class Timeout(Tag):
    pass


@dataclasses.dataclass(frozen=True)
class Skipped(Tag):
    pass


@dataclasses.dataclass
class Result:
    responses_per_node: dict[int, dict[str, Union[Tag, "Access_1.Response"]]] = dataclasses.field(default_factory=dict)
    """
    Keys of the innermost dict are always the same as those in the directive regardless of success.
    Processing of a node stops at first timeout, further items set to Skipped.
    Empty value means that there is no such register.
    """


async def do_calls(
    local_node: "pycyphal.application.Node",
    progress: Callable[[str], None],
    *,
    timeout: float,
    directive: Directive,
) -> Result:
    from uavcan.register import Access_1

    out = Result()
    for node_id, node_dir in directive.registers_per_node.items():
        cln = local_node.make_client(Access_1, node_id)
        try:
            cln.response_timeout = timeout
            responses: dict[str, Access_1.Response | Tag] = {k: Skipped() for k in node_dir}
            for idx, (reg_name, reg_dir) in enumerate(node_dir.items()):
                progress(f"{node_id: 5}: {reg_name!r}")
                resp = await _process_one(cln, reg_name, reg_dir)
                responses[reg_name] = resp
                _logger.info("Result for %r@%r (%r/%r): %r", reg_name, node_id, idx + 1, len(node_dir), resp)
                if isinstance(resp, Timeout):
                    break
        finally:
            cln.close()
        out.responses_per_node[node_id] = responses

    assert out.responses_per_node.keys() == directive.registers_per_node.keys()
    assert all(
        out.responses_per_node[node_id].keys() == directive.registers_per_node[node_id].keys()
        for node_id in directive.registers_per_node
    )
    return out


async def _process_one(
    client: pycyphal.presentation.Client["Access_1"],
    register_name: str,
    directive: RegisterDirective,
) -> Union[TypeCoercionFailure, Timeout, "Access_1.Response"]:
    from pycyphal.application.register import Value
    from uavcan.register import Access_1, Name_1

    # Construct the request object. If we are given the full type information then we can do the job in one query.
    req = Access_1.Request(name=Name_1(register_name), value=directive if isinstance(directive, Value) else None)

    # Go through the network. Empty response means there is no such register, no point proceeding.
    resp = await client(req)
    if resp is None:
        return Timeout()
    assert isinstance(resp, Access_1.Response)
    if resp.value.empty or isinstance(directive, Value):
        return resp

    # Perform type coercion to the discovered type.
    assert callable(directive)
    coerced = directive(resp.value)
    if coerced is None:
        return TypeCoercionFailure(f"Value not coercible to {resp.value}")

    # Send the write request with the updated coerced value.
    req = Access_1.Request(name=Name_1(register_name), value=coerced)
    assert isinstance(req.value, Value)
    return await client(req) or Timeout()


_logger = yakut.get_logger(__name__)



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import time
import itertools
from typing import List, Iterable
import threading
import click
import yakut
from yakut.controller import list_controllers, Controller


_logger = yakut.get_logger(__name__)


@yakut.subcommand(aliases="joy")
def joystick() -> None:
    """
    Show the state of all joysticks and MIDI controllers connected to this computer, htop-style.
    The values printed by this command can be used with the publisher command to alter published messages in real-time.
    This is convenient for controlling various processes or mechanisms in real time during simulation,
    hardware testing, manual operation, etc.

    There is an index associated with each controller that can be referred to along with a specific axis/button name
    from the publisher command (read its docs for more information and usage examples).
    There is also a virtual controller named "null" always available at index 0 with no axes/buttons;
    all of its channels always read out as zero/false.

    X-Box controllers may require a special initialization sequence under GNU/Linux:
    https://gist.github.com/pavel-kirienko/86b9d039151405451130a0fb3896887c
    """
    controllers = [factory() for _name, factory in list_controllers()]
    _logger.info("Using %d controllers: %s", len(controllers), controllers)
    try:
        _run(controllers)
    finally:
        for ctl in controllers:
            ctl.close()


def _run(controllers: List[Controller]) -> None:
    update = threading.Event()
    for ctl in controllers:
        ctl.set_update_hook(update.set)

    for update_count in itertools.count():
        time.sleep(0.05)
        update.wait(1.0)
        update.clear()

        lines = "\n".join(_render_all(controllers))
        click.clear()
        click.echo(lines)
        _logger.debug("%d updates", update_count)


def _render_all(controllers: List[Controller]) -> Iterable[str]:
    yield "Legend: A -- analog axis, B -- push button, T -- toggle switch"
    for idx, ctl in enumerate(controllers):
        yield ""
        yield click.style(f"{idx} ", fg="bright_cyan") + click.style(ctl.name, fg="bright_white", bg="blue")
        sample = ctl.sample()

        if sample.axis:
            yield " ".join(
                click.style(f"A({idx},{axis})=", fg="green") + click.style(f"{value:+.2f}", fg="bright_white")
                for axis, value in sorted(sample.axis.items())
            )
        else:
            yield click.style("No analog axes detected (try moving the controls)", fg="bright_red")

        if sample.button:
            yield " ".join(
                click.style(f"B({idx},{axis})=", fg="cyan") + click.style(f"{value:d}", fg="bright_white")
                for axis, value in sorted(sample.button.items())
            )
        else:
            yield click.style("No buttons detected (try pushing them)", fg="bright_red")

        if sample.toggle:
            yield " ".join(
                click.style(f"T({idx},{axis})=", fg="yellow") + click.style(f"{value:d}", fg="bright_white")
                for axis, value in sorted(sample.toggle.items())
            )
        else:
            yield click.style("No toggles detected (try switching them)", fg="bright_red")



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
import math
from typing import Any, Sequence, TYPE_CHECKING, Callable, Iterable
import logging
from functools import lru_cache
import click
import pydsdl
import pycyphal
from pycyphal.transport import TransferFrom
from pycyphal.presentation import Subscriber
import yakut
from yakut.param.formatter import Formatter
from yakut.util import convert_transfer_metadata_to_builtin
from yakut.subject_specifier_processor import process_subject_specifier, SubjectResolver
from ._sync import Synchronizer, SynchronizerFactory

if TYPE_CHECKING:
    import pycyphal.application

_logger = yakut.get_logger(__name__)

SYNC_MONOCLUST_TOLERANCE_MINMAX_TS_FIELD = 1e-6, 10.0
"""
Assume that externally provided timestamp is accurate at least for high-frequency topics.
"""

SYNC_MONOCLUST_TOLERANCE_MINMAX_TS_ARRIVAL = 0.02, 60.0
"""
A large minimum is needed due to low arrival timestamping accuracy.
This may be especially bad on Windows where the timestamping resolution may be as low as ~16 ms,
so the minimum tolerance should not be lower than that.
"""


class Config:
    def __init__(self) -> None:
        self._synchronizer_factory: SynchronizerFactory | None = None

    def set_synchronizer_factory(self, val: SynchronizerFactory) -> None:
        if self._synchronizer_factory is not None:
            raise click.UsageError("Cannot use more than one synchronizer")
        self._synchronizer_factory = val
        _logger.debug("Using synchronizer factory: %r", val)

    def get_synchronizer_factory(self) -> SynchronizerFactory:
        if self._synchronizer_factory is None:
            from ._sync_async import make_sync_async

            self.set_synchronizer_factory(make_sync_async)
        assert self._synchronizer_factory is not None
        return self._synchronizer_factory


def _has_field(model: pydsdl.CompositeType, name: str, type_full_name: str) -> bool:
    for field in model.fields:
        dt = field.data_type
        if field.name == name and isinstance(dt, pydsdl.CompositeType) and dt.full_name == type_full_name:
            return True
    return False


def _ensure_timestamp_field_synchronization_is_possible(model: pydsdl.CompositeType) -> None:
    if not _has_field(model, "timestamp", "uavcan.time.SynchronizedTimestamp"):
        raise click.ClickException(
            f"Synchronization on timestamp field is not possible for {model} because there is no such field. "
            f"Please use another synchronization policy or use only timestamped data types."
        )


def _handle_option_synchronizer_monoclust_timestamp_field(
    ctx: click.Context,
    _param: click.Parameter,
    value: float | None,
) -> None:
    if value is not None:
        from pycyphal.presentation.subscription_synchronizer import get_timestamp_field
        from ._sync_monoclust import make_sync_monoclust

        value = float(value)
        tol = SYNC_MONOCLUST_TOLERANCE_MINMAX_TS_FIELD if not math.isfinite(value) else (value, value)
        _logger.debug("Configuring field timestamp monoclust synchronizer with tolerance=%r", tol)

        def fac(subs: Iterable[Subscriber[Any]]) -> Synchronizer:
            subs = list(subs)
            for s in subs:
                _ensure_timestamp_field_synchronization_is_possible(pycyphal.dsdl.get_model(s.dtype))
            return make_sync_monoclust(subs, f_key=get_timestamp_field, tolerance_minmax=tol)

        ctx.ensure_object(Config).set_synchronizer_factory(fac)


def _handle_option_synchronizer_monoclust_timestamp_arrival(
    ctx: click.Context,
    _param: click.Parameter,
    value: float | None,
) -> None:
    if value is not None:
        from pycyphal.presentation.subscription_synchronizer import get_local_reception_timestamp
        from ._sync_monoclust import make_sync_monoclust

        value = float(value)
        tol = SYNC_MONOCLUST_TOLERANCE_MINMAX_TS_ARRIVAL if not math.isfinite(value) else (value, value)
        _logger.debug("Configuring arrival timestamp monoclust synchronizer; tolerance=%r", tol)
        ctx.ensure_object(Config).set_synchronizer_factory(
            lambda subs: make_sync_monoclust(subs, f_key=get_local_reception_timestamp, tolerance_minmax=tol)
        )


def _handle_option_synchronizer_transfer_id(ctx: click.Context, _param: click.Parameter, value: bool) -> None:
    if value:
        from ._sync_transfer_id import make_sync_transfer_id

        _logger.debug("Configuring transfer-ID synchronizer")
        ctx.ensure_object(Config).set_synchronizer_factory(make_sync_transfer_id)


@yakut.subcommand(aliases=["sub", "s"])
@click.argument("subject", type=str, nargs=-1)
@click.option(
    "--with-metadata/--no-metadata",
    "+M/-M",
    default=False,
    show_default=True,
    help="When enabled, each message object is prepended with an extra field named `_meta_`.",
)
@click.option(
    "--count",
    "-N",
    type=int,
    metavar="CARDINAL",
    help=f"""
Exit automatically after this many messages (or synchronous message groups) have been received. No limit by default.
""",
)
@click.option(
    "--redraw",
    "--no-scroll",
    "-R",
    is_flag=True,
    help="Clear terminal before printing output. This option only has effect if stdout is a tty.",
)
@click.option(
    "--sync-monoclust-field",
    "--smcf",
    callback=_handle_option_synchronizer_monoclust_timestamp_field,
    expose_value=False,
    type=float,
    is_flag=False,
    flag_value=float("nan"),
    help=f"""
Use the monotonic clustering synchronizer with the message timestamp field as the clustering key.
All data types shall be timestamped for this to work.
The optional value is the synchronization tolerance in seconds; autodetect if not specified.
""",
)
@click.option(
    "--sync-monoclust-arrival",
    "--smca",
    "--sync",  # This is currently the default because it works with any data type.
    callback=_handle_option_synchronizer_monoclust_timestamp_arrival,
    expose_value=False,
    type=float,
    is_flag=False,
    flag_value=float("nan"),
    help=f"""
Use the monotonic clustering synchronizer with the local arrival timestamp as the clustering key.
Works with all data types but may perform poorly depending on the local timestamping accuracy and system latency.
The optional value is the synchronization tolerance in seconds; autodetect if not specified.
""",
)
@click.option(
    "--sync-transfer-id",
    "--stid",
    callback=_handle_option_synchronizer_transfer_id,
    expose_value=False,
    is_flag=True,
    help=f"""
Use the transfer-ID synchronizer.
Messages that originate from the same node AND share the same transfer-ID will be grouped together.
""",
)
@yakut.pass_purser
@yakut.asynchronous(interrupted_ok=True)
async def subscribe(
    purser: yakut.Purser,
    subject: tuple[str, ...],
    with_metadata: bool,
    count: int | None,
    redraw: bool,
) -> None:
    """
    Subscribe to specified subjects and print messages to stdout.
    It is recommended to make this node anonymous (i.e., without a node-ID)
    to avoid additional traffic on the network from this tool.

    The arguments are a list of message data type names prepended with the subject-ID;
    the subject-ID may be omitted if the data type defines a fixed one;
    or the type can be omitted to engage automatic type discovery
    (discovery may fail if the local node is anonymous as it will be unable to issue RPC-service requests).
    The short data type name is case-insensitive for convenience.
    The accepted forms are:

    \b
        SUBJECT_ID:TYPE_NAME[.MAJOR[.MINOR]]
        TYPE_NAME[.MAJOR[.MINOR]]
        SUBJECT_ID

    The output documents (YAML/JSON/etc depending on the chosen format) will contain one message object each,
    unless synchronization is enabled via --sync-*.
    In that case, each output document will contain a complete synchronized group of messages (ordering retained).
    Synchronous subscription is intended for subscribing to a group of coupled subjects like coupled sensor feeds.
    More on subscription synchronization is available in the PyCyphal docs.

    Each received message or synchronized group is emitted to stdout as a key-value mapping,
    where the keys are subject-IDs and values are the received message objects.

    Examples:

    \b
        yakut sub 33:uavcan.si.unit.angle.scalar --with-metadata --count=1
        yakut sub 33 42 5789 --sync-monoclust-arrival=0.1
        yakut sub uavcan.node.heartbeat
        y sub 1220 1230 1240 1250 --sync-monoclust

    Extracting a specific field, sub-object, data manipulation:

    \b
        y sub uavcan.node.heartbeat | jq '.[].health.value'
        y sub uavcan.node.heartbeat +M | jq '.[]|[._meta_.transfer_id, ._meta_.timestamp.system]'
    """
    config = click.get_current_context().ensure_object(Config)
    try:
        from pycyphal.application import Node
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.ClickException(make_usage_suggestion(ex.name))

    finalizers: list[Callable[[], None]] = []
    try:
        _logger.debug("subject=%r, with_metadata=%r, count=%r", subject, with_metadata, count)
        if count is not None and count <= 0:
            _logger.warning("Nothing to do because count=%s", count)
            return

        count = count if count is not None else sys.maxsize
        formatter = purser.make_formatter()

        # Node construction should be delayed as much as possible to avoid unnecessary interference
        # with the bus and hardware. This is why we use the factory here instead of constructing the node eagerly.
        @lru_cache(None)
        def get_node() -> Node:
            node = purser.get_node("subscribe", allow_anonymous=True)
            finalizers.append(node.close)
            return node

        subscribers: list[Subscriber[Any]] = await _make_subscribers(subject, get_node)
        finalizers += [s.close for s in subscribers]
        if len(subscribers) == 0:
            _logger.warning("Nothing to do because no subjects are specified")
            return
        synchronizer = config.get_synchronizer_factory()(subscribers)

        # The node is closed through the finalizers at exit.
        # Note that we can't close the node before closing the subscribers to avoid resource errors inside PyCyphal.
        node = get_node()
        if node.id is not None:
            _logger.info("It is recommended to use an anonymous node with this command")
        node.start()
        try:
            await _run(synchronizer, formatter, with_metadata=with_metadata, count=count, redraw=redraw)
        finally:
            if _logger.isEnabledFor(logging.INFO):
                _logger.info("%s", node.presentation.transport.sample_statistics())
                for sub in subscribers:
                    _logger.info("% 4s: %s", sub.port_id, sub.sample_statistics())
    finally:
        pycyphal.util.broadcast(finalizers[::-1])()


async def _make_subscribers(
    specifiers: Sequence[str],
    node_provider: Callable[[], "pycyphal.application.Node"],
) -> list[Subscriber[Any]]:
    subject_resolver: SubjectResolver | None = None

    def get_resolver() -> SubjectResolver:
        nonlocal subject_resolver
        if subject_resolver is None:
            node = node_provider()
            if node.id is None:
                raise click.ClickException(
                    f"Cannot use automatic discovery because the local node is anonymous, "
                    f"so it cannot access the introspection services on remote nodes. "
                    f"You need to either fully specify the subjects explicitly or assign a local node-ID."
                )
            subject_resolver = SubjectResolver(node)
        return subject_resolver

    try:
        id_types = [await process_subject_specifier(ds, get_resolver) for ds in specifiers]
        # Construct the node only after we have verified that the specifiers are valid and dtypes/ids are resolved.
        return [node_provider().make_subscriber(dtype, subject_id) for subject_id, dtype in id_types]
    finally:
        if subject_resolver:
            subject_resolver.close()


async def _run(synchronizer: Synchronizer, formatter: Formatter, with_metadata: bool, count: int, redraw: bool) -> None:
    def process_group(group: tuple[tuple[tuple[Any, TransferFrom] | None, Subscriber[Any]], ...]) -> None:
        nonlocal count
        outer: dict[int, dict[str, Any]] = {}
        for maybe_msg_meta, subscriber in group:
            if maybe_msg_meta is None:
                continue  # Asynchronous mode.
            msg, meta = maybe_msg_meta
            assert isinstance(meta, TransferFrom) and isinstance(subscriber, Subscriber)
            bi: dict[str, Any] = {}  # We use updates to ensure proper dict ordering: metadata before data
            if with_metadata:
                bi.update(convert_transfer_metadata_to_builtin(meta, dtype=subscriber.dtype))
            bi.update(pycyphal.dsdl.to_builtin(msg))
            outer[subscriber.port_id] = bi

        if redraw:
            click.clear()
        sys.stdout.write(formatter(outer))
        sys.stdout.flush()
        count -= 1
        if count <= 0:
            _logger.debug("Reached the specified synchronized group count, stopping")
            raise _Break

    try:
        await synchronizer(process_group)
    except _Break:
        pass


class _Break(Exception):
    pass



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
from typing import Any, Callable, TYPE_CHECKING
import decimal
from functools import lru_cache
import click
import pycyphal
import yakut
from yakut.enum_param import EnumParam
from yakut.param.formatter import Formatter, FormatterHints
from yakut.util import convert_transfer_metadata_to_builtin
from yakut import dtype_loader

if TYPE_CHECKING:
    from pycyphal.application import Node


_RESOLVER_MIN_RESPONSE_TIMEOUT = 1.0

_logger = yakut.get_logger(__name__)


def _validate_request_fields(ctx: click.Context, param: click.Parameter, value: str) -> Any:
    from yakut.yaml import EvaluableLoader

    eval_context: dict[str, Any] = {}  # Add useful members later.
    try:
        fields = EvaluableLoader(eval_context).load(value)
    except Exception as ex:
        raise click.BadParameter(f"Could not parse the request object fields: {ex}", ctx=ctx, param=param)
    return fields


@yakut.subcommand(aliases="q")
@click.argument("server_node_id", metavar="SERVER_NODE_ID", type=int, required=True)
@click.argument("service", metavar="SERVICE", type=str, required=True)
@click.argument("request_fields", metavar="FIELDS", type=str, callback=_validate_request_fields, default="{}")
@click.option(
    "--timeout",
    "-T",
    type=float,
    default=pycyphal.presentation.DEFAULT_SERVICE_REQUEST_TIMEOUT,
    show_default=True,
    metavar="SECONDS",
    help=f"Request timeout; how long to wait for the response before giving up.",
)
@click.option(
    "--priority",
    "-P",
    default=pycyphal.presentation.DEFAULT_PRIORITY,
    type=EnumParam(pycyphal.transport.Priority),
    help=f"Priority of the request transfer. [default: {pycyphal.presentation.DEFAULT_PRIORITY.name}]",
)
@click.option(
    "--with-metadata/--no-metadata",
    "+M/-M",
    default=False,
    show_default=True,
    help="When enabled, the response object is prepended with an extra field named `_meta_`.",
)
@yakut.pass_purser
@yakut.asynchronous()
async def call(
    purser: yakut.Purser,
    server_node_id: int,
    service: str,
    request_fields: Any,
    timeout: float,
    priority: pycyphal.transport.Priority,
    with_metadata: bool,
) -> None:
    """
    Invoke an RPC-service using the specified request object and print the response.
    The local node will also publish on standard subjects like Heartbeat and provide some standard RPC-services.

    The first positional argument is the server node-ID.

    The second positional argument specifies which service to invoke on the server, which can be done in two ways:
    1. specify the numerical service-ID and its type separated by colon;
    2. specify the name of the service only and let Yakut retrieve the type information from the server.
    The syntax of the available options is as follows:

    \b
        [SERVICE_ID:]TYPE_NAME[.MAJOR[.MINOR]]
        SERVICE_NAME[:TYPE_NAME[.MAJOR[.MINOR]]]

    The short data type name is case-insensitive for convenience.
    In the data type name, "/" or "\\" can be used instead of "." for convenience and filesystem autocompletion.
    If the data type is specified without the service-ID, a fixed service-ID shall be defined for this data type.
    Missing version numbers default to the latest available.
    The two formats are differentiated automatically using heuristics.

    The third positional argument specifies the values of the request object fields in YAML format
    (or JSON, which is a subset of YAML).
    Missing fields will be left at their default values.
    If omitted, this argument defaults to an empty object: `{}`.
    For more info about the format see pycyphal.dsdl.update_from_builtin().

    The output will be printed as a key-value mapping of one element where the key is the service-ID
    and the value is the received response object (optionally with metadata).

    Examples:

    \b
        yakut call 42 uavcan.node.getinfo +M -T3 -Pe
        yakut call 42 least_squares 'points: [{x: 10, y: 1}, {x: 20, y: 2}]'
        yakut call 42 least_squares:sirius_cyber_corp.PerformLinearLeastSquaresFit '[[10, 1], [20, 2]]'
    """
    try:
        from pycyphal.application import Node
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.ClickException(make_usage_suggestion(ex.name))

    _logger.debug(
        "server_node_id=%s, service=%r, request_fields=%r, timeout=%.6f, priority=%s, with_metadata=%s",
        server_node_id,
        service,
        request_fields,
        timeout,
        priority,
        with_metadata,
    )
    finalizers: list[Callable[[], None]] = []
    try:
        formatter = purser.make_formatter(FormatterHints(single_document=True))

        # The cached factory is needed to postpone node initialization as much as possible because it disturbs
        # the network and the networking hardware (if any) and is usually costly.
        @lru_cache(None)
        def get_node() -> Node:
            node = purser.get_node("call", allow_anonymous=False)
            finalizers.append(node.close)
            return node

        service_id, dtype = await _resolve(
            service,
            server_node_id,
            get_node,
            response_timeout=max(_RESOLVER_MIN_RESPONSE_TIMEOUT, timeout),
        )
        if not pycyphal.dsdl.is_service_type(dtype):
            raise click.ClickException(f"{pycyphal.dsdl.get_model(dtype)} is not a service type") from None
        request = pycyphal.dsdl.update_from_builtin(dtype.Request(), request_fields)
        _logger.info("Request object: %r", request)

        client = get_node().make_client(dtype, server_node_id, service_id)
        finalizers.append(client.close)
        client.response_timeout = timeout
        client.priority = priority

        get_node().start()
        await _run(client, request, formatter, with_metadata=with_metadata)
    finally:
        pycyphal.util.broadcast(finalizers[::-1])()


async def _run(
    client: pycyphal.presentation.Client[Any],
    request: Any,
    formatter: Formatter,
    with_metadata: bool,
) -> None:
    request_ts_transport: pycyphal.transport.Timestamp | None = None

    def on_transfer_feedback(fb: pycyphal.transport.Feedback) -> None:
        nonlocal request_ts_transport
        request_ts_transport = fb.first_frame_transmission_timestamp

    client.output_transport_session.enable_feedback(on_transfer_feedback)

    request_ts_application = pycyphal.transport.Timestamp.now()
    result = await client.call(request)
    response_ts_application = pycyphal.transport.Timestamp.now()
    if result is None:  # TODO this should exit with yakut.util.EXIT_CODE_UNSUCCESSFUL
        raise click.ClickException(f"The request has timed out after {client.response_timeout:0.1f} seconds")
    if not request_ts_transport:  # pragma: no cover
        request_ts_transport = request_ts_application
        _logger.warning(
            "The transport implementation is misbehaving: feedback was never emitted; "
            "falling back to software timestamping. "
            "Please submit a bug report. Involved instances: client=%r, result=%r",
            client,
            result,
        )
    response, transfer = result
    transport_duration = transfer.timestamp.monotonic - request_ts_transport.monotonic
    application_duration = response_ts_application.monotonic - request_ts_application.monotonic
    _logger.info(
        "Request duration [second]: "
        "transport layer: %.6f, application layer: %.6f, application layer overhead: %.6f",
        transport_duration,
        application_duration,
        application_duration - transport_duration,
    )

    bi: dict[str, Any] = {}  # We use updates to ensure proper dict ordering: metadata before data
    if with_metadata:
        qnt = decimal.Decimal("0.000001")
        bi.update(
            convert_transfer_metadata_to_builtin(
                transfer,
                dtype=client.dtype,
                rtt=(transfer.timestamp.monotonic - request_ts_transport.monotonic).quantize(qnt),
            )
        )
    bi.update(pycyphal.dsdl.to_builtin(response))

    sys.stdout.write(formatter({client.port_id: bi}))
    sys.stdout.flush()


async def _resolve(
    raw_spec: str,
    server_node_id: int,
    node_provider: Callable[[], Node],
    *,
    response_timeout: float,
) -> tuple[int, Any]:
    """
    Decentralized discovery: if service-ID and type are given by the user, simply load the specified type and return.
    Otherwise we have to query the server node to find out the type and/or ID of the service.
    """
    specs = raw_spec.split(":")
    if not 1 <= len(specs) <= 2:
        click.BadParameter(f"Service specifier invalid: {raw_spec!r}")

    if len(specs) == 2:
        dtype = dtype_loader.load_dtype(specs[1])
        try:
            return int(specs[0]), dtype
        except ValueError:
            pass
        _logger.info(
            "Querying server node %r for service-ID information; locally specified type is %r", server_node_id, dtype
        )
        resolved = await _resolve_service_id_type(
            node_provider().presentation,
            specs[0],
            server_node_id,
            response_timeout=response_timeout,
        )
        if not resolved:
            raise click.ClickException(
                f"Could not resolve service {specs[0]!r} via node {server_node_id}. "
                f"The remote node might be offline or it may not support automatic discovery. "
                f"Consider specifying the service-ID and the data type explicitly?"
            )
        service_id, dtype_name_remote = resolved
        _logger.debug("Using locally specified type %r; server uses %r", dtype, dtype_name_remote)
        return service_id, dtype

    (ty_or_srv,) = specs
    del specs
    possibly_dtype_name = ty_or_srv.count(".") >= 1
    if possibly_dtype_name:
        try:
            dtype = dtype_loader.load_dtype(ty_or_srv)
        except dtype_loader.LoadError:
            pass
        else:
            fpid = pycyphal.dsdl.get_fixed_port_id(dtype)
            if fpid is None:
                raise click.ClickException(f"Type {pycyphal.dsdl.get_model(dtype)} does not have a fixed port-ID")
            _logger.info("Using type %r with its fixed port-ID %r (network discovery not required)", dtype, fpid)
            return fpid, dtype

    _logger.info("Heuristic: %r does not appear to be a type name, assuming it to be a port name", ty_or_srv)
    resolved = await _resolve_service_id_type(
        node_provider().presentation,
        ty_or_srv,
        server_node_id,
        response_timeout=response_timeout,
    )
    if not resolved:
        raise click.ClickException(
            f"Could not resolve service {ty_or_srv!r} via node {server_node_id}. "
            f"The remote node might be offline or it may not support automatic discovery. "
            f"Consider specifying the service-ID and the data type explicitly?"
        )
    service_id, dtype_name = resolved
    _logger.info("Resolved from server %r: id=%r dtype=%r", server_node_id, service_id, dtype_name)
    if dtype_name is None:
        raise click.ClickException(
            f"Remote node {server_node_id} does not provide data type information for service {ty_or_srv}, "
            f"nor is the type specified locally."
        )
    # When the type is obtained from remote node, allow the minor version to be different.
    return service_id, dtype_loader.load_dtype(dtype_name, allow_minor_version_mismatch=True)


async def _resolve_service_id_type(
    pres: pycyphal.presentation.Presentation,
    port_name: str,
    server_node_id: int,
    *,
    response_timeout: float,
) -> tuple[int, str | None] | None:
    from pycyphal.application.register import ValueProxy
    from uavcan.register import Access_1, Name_1

    c_access = pres.make_client_with_fixed_service_id(Access_1, server_node_id)
    try:
        c_access.response_timeout = response_timeout
        req = Access_1.Request(name=Name_1(f"uavcan.srv.{port_name}.id"))
        resp = await c_access(req)
        if resp is None:
            _logger.info("Request to %r has timed out: %s", server_node_id, req)
            return None
        assert isinstance(resp, Access_1.Response)
        port_id = int(ValueProxy(resp.value))
        if not 0 <= port_id <= pycyphal.transport.ServiceDataSpecifier.SERVICE_ID_MASK:
            _logger.debug("Service %r at node %r is not configured", port_name, server_node_id)
            return None

        req = Access_1.Request(name=Name_1(f"uavcan.srv.{port_name}.type"))
        resp = await c_access(req)
        if resp is None:
            _logger.info("Request to %r has timed out: %s", server_node_id, req)
            return None
        assert isinstance(resp, Access_1.Response)
        dtype_name = str(ValueProxy(resp.value)) if resp.value.string else None
    finally:
        c_access.close()
    return port_id, dtype_name



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable, Optional
import logging
import pycyphal
from yakut.util import METADATA_KEY

if TYPE_CHECKING:
    import pycyphal.application
    from pycyphal.application.register import Value
    from uavcan.register import Access_1


async def fetch_registers(
    presentation: pycyphal.presentation.Presentation,
    node_id: int,
    *,
    predicate: Callable[[str], bool] = lambda *_: True,
    timeout: float = pycyphal.presentation.DEFAULT_SERVICE_REQUEST_TIMEOUT,
    priority: pycyphal.transport.Priority = pycyphal.transport.Priority.LOW,
) -> dict[str, "pycyphal.application.register.ValueProxy"] | None:
    """
    Obtain registers from the specified remote node for whose names the predicate is true.
    Returns None on network timeout.
    """
    from pycyphal.application.register import ValueProxy as RegisterValue
    from uavcan.register import Access_1, List_1, Name_1

    # Fetch register names.
    c_list = presentation.make_client_with_fixed_service_id(List_1, node_id)
    c_list.response_timeout = timeout
    c_list.priority = priority
    names: list[str] = []
    while True:
        req: Any = List_1.Request(len(names))
        resp = await c_list(req)
        if resp is None:
            _logger.warning("Request to %r has timed out: %s", node_id, req)
            return None
        assert isinstance(resp, List_1.Response)
        if not resp.name.name.tobytes():
            break
        names.append(resp.name.name.tobytes().decode())
    _logger.debug("Register names fetched from node %r: %s", node_id, names)
    c_list.close()
    del c_list

    names = list(filter(predicate, names))

    # Then fetch the registers themselves.
    c_access = presentation.make_client_with_fixed_service_id(Access_1, node_id)
    c_access.response_timeout = timeout
    c_access.priority = priority
    regs: dict[str, RegisterValue] = {}
    for nm in names:
        req = Access_1.Request(name=Name_1(nm))
        resp = await c_access(req)
        if resp is None:
            _logger.warning("Request to %r has timed out: %s", node_id, req)
            return None
        assert isinstance(resp, Access_1.Response)
        regs[nm] = RegisterValue(resp.value)
    c_access.close()

    return regs


def unexplode_value(xpl: Any, prototype: Optional["Value"] = None) -> Optional["Value"]:
    """
    Reverse the effect of :func:`explode`.
    Returns None if the exploded form is invalid or not applicable to the prototype.
    Some simplified exploded forms can be unexploded only if the prototype
    is given because simplification erases type information.
    Some unambiguous simplified forms may be unexploded autonomously.

    >>> from tests.dsdl import ensure_compiled_dsdl
    >>> ensure_compiled_dsdl()
    >>> from pycyphal.application.register import Value, Natural16
    >>> ux = unexplode_value

    >>> ux(None)                                         # None is a simplified form of Empty.
    uavcan.register.Value...(empty=...)
    >>> ux({"integer8": {"value": [1,2,3]}, "_meta_": {"whatever": 0}})  # Metadata ignored.
    uavcan.register.Value...(integer8=...[1,2,3]))
    >>> ux({"integer8": {"value": [1,2,3]}})             # Pure Value (same as above)
    uavcan.register.Value...(integer8=...[1,2,3]))
    >>> ux([1,2,3]) is None                              # Prototype required.
    True
    >>> ux([1,2,3], Value(natural16=Natural16([0,0,0])))
    uavcan.register.Value...(natural16=...[1,2,3]))
    >>> ux(123, Value(natural16=Natural16([0])))
    uavcan.register.Value...(natural16=...[123]))
    >>> ux("abc", Value(natural16=Natural16([0]))) is None # Not applicable
    True

    Roundtrip:

    >>> unexplode_value(explode_value(Value(natural16=Natural16([0,1,2])), metadata={"a": 654}))
    uavcan.register.Value...(natural16=...[0,1,2]))
    """
    from pycyphal.dsdl import update_from_builtin
    from pycyphal.application.register import ValueProxy, Value, ValueConversionError

    if xpl is None:
        return Value()
    if isinstance(xpl, dict) and xpl:  # Empty dict is not a valid representation.
        try:
            res = update_from_builtin(
                Value(),
                {k: v for k, v in xpl.items() if k.strip("_") == k},  # Strip metadata fields.
            )
            assert isinstance(res, Value)
            return res
        except (ValueError, TypeError):
            pass
    if prototype is not None:
        ret = ValueProxy(prototype)
        try:
            ret.assign(xpl)
            assert isinstance(ret.value, Value)
            return ret.value
        except ValueConversionError:
            pass
    return None


def explode_value(val: "Value", *, simplify: bool = False, metadata: dict[str, Any] | None = None) -> Any:
    """
    Represent the register value using primitives (list, dict, string, etc.).
    If simplified mode is selected,
    the metadata and type information will be discarded and only a human-friendly representation of the
    value will be constructed.
    The reconstruction back to the original form is a bit involved but we provide :func:`unexplode` for that.
    The metadata is added under a key ``_meta_``, if there is any, but it is ignored in simplified mode.
    """
    if not simplify:
        out = pycyphal.dsdl.to_builtin(val)
        if metadata is not None:
            out[METADATA_KEY] = dict(metadata)
        return out
    return _simplify_value(val)


def _simplify_value(msg: "Value") -> Any:
    """
    Construct simplified human-friendly representation of the register value using primitives (list, string, etc.).
    Designed for use with commands that output compact register values in YAML/JSON/TSV/whatever,
    discarding the detailed type information.

    >>> from tests.dsdl import ensure_compiled_dsdl
    >>> ensure_compiled_dsdl()
    >>> from pycyphal.application.register import Value, Empty
    >>> from pycyphal.application.register import Integer8, Natural8, Integer32, String, Unstructured

    >>> None is _simplify_value(Value())  # empty is none
    True
    >>> _simplify_value(Value(integer8=Integer8([123])))
    123
    >>> _simplify_value(Value(natural8=Natural8([123, 23])))
    [123, 23]
    >>> _simplify_value(Value(integer32=Integer32([123, -23, 105])))
    [123, -23, 105]
    >>> _simplify_value(Value(integer32=Integer32([99999])))
    99999
    >>> _simplify_value(Value(string=String("Hello world")))
    'Hello world'
    >>> _simplify_value(Value(unstructured=Unstructured(b"Hello world")))
    b'Hello world'
    """
    # This is kinda crude, perhaps needs improvement.
    if msg.empty:
        return None
    if msg.unstructured:
        return msg.unstructured.value.tobytes()
    if msg.string:
        return msg.string.value.tobytes().decode(errors="replace")
    ((_ty, val),) = pycyphal.dsdl.to_builtin(msg).items()
    val = val["value"]
    val = list(val.encode() if isinstance(val, str) else val)
    if len(val) == 1:  # One-element arrays shown as scalars.
        (val,) = val
    return val


def get_access_response_metadata(val: "Access_1.Response") -> dict[str, Any]:
    """
    This is for use with :func:`explode_value`.
    """
    return {
        "mutable": val.mutable,
        "persistent": val.persistent,
    }


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
import dataclasses
import logging
from typing import Callable, Any, cast
from collections.abc import Mapping, Collection
import click


@dataclasses.dataclass(frozen=True)
class FormatterHints:
    single_document: bool = False
    """
    If true, document separators will not be emitted (if applicable).
    """


Formatter = Callable[[Any], str]
"""
The output represents a complete document with the terminating newline if such is necessary.
The caller need not add anything.
If sending this to stdout, do not use print() as-is because it will add an extra newline at the end.
"""

FormatterFactory = Callable[[FormatterHints], Formatter]


def formatter_factory_option(f: Callable[..., Any]) -> Callable[..., Any]:
    override: str | None = None

    def install_override(_ctx: click.Context, _param: click.Parameter, value: str) -> None:
        nonlocal override
        override = value or override

    def validate(ctx: click.Context, param: click.Parameter, value: str) -> FormatterFactory:
        try:
            return _FORMATTERS[(override or value).upper()]
        except LookupError:
            raise click.BadParameter(f"Invalid format name: {value!r}", ctx=ctx, param=param) from None

    doc = f"""
The format of data printed to stdout.
This option is only relevant for commands that produce data (sub, call, etc.).

The final representation of the output data is constructed from an intermediate "builtin-based" representation,
which is a simplified form that is stripped of the detailed DSDL type information, like JSON.
For more info please read the PyCyphal documentation on builtin-based representations.

Option "auto" (default) selects YAML if the output is shown on the terminal for the benefit of the user;
if the output is redirected (e.g., piped to another command or to a file),
JSON is selected to enable compatibility with jq and other 3rd-party stream processing tools.
It helps to remember that JSON is a subset of YAML.

YAML separates objects with `---`.
JSON and TSV (tab separated values) keep exactly one object per line.

TSV is intended for use with third-party software
such as computer algebra systems or spreadsheet processors.

TSVH is just TSV with the header included.

TSVFC is TSVH with extra column for curly braces and square brackets. These are extra format columns that help the
reader understand the structure of the data without looking at the headers.
"""
    f = click.option(
        "--format",
        "-F",
        "formatter_factory",
        envvar="YAKUT_FORMAT",
        type=click.Choice(list(_FORMATTERS.keys()), case_sensitive=False),
        callback=validate,
        default=list(_FORMATTERS.keys())[0],
        show_default=True,
        help=doc,
    )(f)

    def shortcut(opt: str) -> None:
        nonlocal f
        f = click.option(
            f"--{opt}",
            "-" + opt[0],
            flag_value=opt,
            callback=install_override,
            help=f"Same as --format={opt}",
            is_eager=True,
            expose_value=False,
        )(f)

    shortcut("yaml")
    shortcut("json")
    shortcut("tsvh")
    return f


def _make_yaml_formatter(hints: FormatterHints) -> Formatter:
    from yakut.yaml import Dumper

    dumper = Dumper(explicit_start=not hints.single_document)
    return dumper.dumps


def _make_json_formatter(_hints: FormatterHints) -> Formatter:
    # We prefer simplejson over the standard json because the native json lacks important capabilities:
    #  - simplejson preserves dict ordering, which is very important for UX.
    #  - simplejson supports Decimal.
    import simplejson as json  # type: ignore

    return lambda data: cast(str, json.dumps(data, ensure_ascii=False, separators=(",", ":"))) + _NEWLINE


def _insert_format_specifier(
    items: list[tuple[str, Any]],
    key: str,
    instance: Collection[Any] | Mapping[Any, Any],
    is_start: bool = True,
) -> None:
    is_list = isinstance(instance, Collection) and not isinstance(instance, str)
    is_dictionary = isinstance(instance, Mapping)
    if is_start:
        if is_dictionary:
            items.append((key + "{", "{"))
        elif is_list:
            items.append((key + "[", "["))
    else:
        if is_dictionary:
            items.append((key + "}", "}"))
        elif is_list:
            items.append((key + "]", "]"))


def _flatten_start(
    outer: Any,
    parent_key: str = "",
    sep: str = ".",
    with_format_specifiers: bool = False,
) -> dict[str, Any]:
    def flatten(data: Any, parent_key: str = "") -> dict[str, Any]:
        def add_item(items: list[tuple[str, Any]], new_key: str, v: Mapping[Any, Any] | Collection[Any]) -> None:
            if with_format_specifiers:
                _insert_format_specifier(items, new_key, v)
            if isinstance(v, Mapping) or (isinstance(v, Collection) and not isinstance(v, str)):
                for_extension = flatten(v, new_key)
                if for_extension is not None:
                    # noinspection PyTypeChecker
                    items.extend(for_extension.items())
            else:
                items.append((new_key, v))
            if with_format_specifiers:
                _insert_format_specifier(items, new_key, v, is_start=False)

        if isinstance(data, Mapping):
            items: list[tuple[str, Any]] = []
            for k, v in data.items():
                new_key = parent_key + sep + str(k) if parent_key else str(k)
                add_item(items, new_key, v)
            return dict(items)
        if isinstance(data, Collection) and not isinstance(data, str):
            items = []
            for i, v in enumerate(data):
                new_key = parent_key + sep + f"[{i}]" if parent_key else str(f"[{i}]")
                add_item(items, new_key, v)
            return dict(items)
        return {}

    return flatten(outer, parent_key)


def _make_tsv_formatter(hints: FormatterHints) -> Formatter:
    # TODO: if single_document, transpose the top-level dict to have the keys on the leftmost row.
    # Transpose lists in a similar manner.
    _ = hints  # TODO not used yet

    def tsv_format_function(data: Any) -> str:
        return "\t".join(str(v) for k, v in _flatten_start(data).items()) + _NEWLINE

    return tsv_format_function


def _make_tsvh_formatter_factory(with_format_specifiers: bool) -> FormatterFactory:
    def make_tsvh_formatter(hints: FormatterHints) -> Formatter:
        _ = hints  # TODO see above
        is_first_time = True

        def tsv_format_function_with_header(data: Any) -> str:
            nonlocal is_first_time
            if is_first_time:
                is_first_time = False
                return _NEWLINE.join(
                    (
                        "\t".join(
                            str(k)
                            for k, v in _flatten_start(data, with_format_specifiers=with_format_specifiers).items()
                        ),
                        "\t".join(
                            str(v)
                            for k, v in _flatten_start(data, with_format_specifiers=with_format_specifiers).items()
                        ),
                        "",
                    )
                )
            return (
                "\t".join(
                    str(v) for k, v in _flatten_start(data, with_format_specifiers=with_format_specifiers).items()
                )
                + _NEWLINE
            )

        return tsv_format_function_with_header

    return make_tsvh_formatter


def _make_auto(hints: FormatterHints) -> Formatter:
    fac = _make_yaml_formatter if sys.stdout.isatty() else _make_json_formatter
    _logger.debug("Automatically selected formatter: %r", fac)
    return fac(hints)


_FORMATTERS = {
    "AUTO": _make_auto,
    "YAML": _make_yaml_formatter,
    "JSON": _make_json_formatter,
    "TSV": _make_tsv_formatter,
    "TSVH": _make_tsvh_formatter_factory(with_format_specifiers=False),
    "TSVFC": _make_tsvh_formatter_factory(with_format_specifiers=True),
}

_NEWLINE = "\n"

_logger = logging.getLogger(__name__)


def _unittest_formatter() -> None:
    default_hints = FormatterHints()

    obj = {
        2345: {
            "abc": {
                "def": [123, 456],
            },
            "ghi": 789,
        }
    }
    assert (
        _FORMATTERS["YAML"](default_hints)(obj)
        == """---
2345:
  abc:
    def: [123, 456]
  ghi: 789
"""
    )
    assert _FORMATTERS["JSON"](default_hints)(obj) == '{"2345":{"abc":{"def":[123,456]},"ghi":789}}\n'
    assert _FORMATTERS["TSV"](default_hints)(obj) == "123\t456\t789\n"
    tsvh_formatter = _FORMATTERS["TSVH"](default_hints)
    # first time should include a header
    assert tsvh_formatter(obj) == "2345.abc.def.[0]\t2345.abc.def.[1]\t2345.ghi\n123\t456\t789\n"
    # subsequent calls shouldn't include a header
    assert tsvh_formatter(obj) == "123\t456\t789\n"
    from decimal import Decimal
    from math import nan

    obj = {
        142: {
            "_meta_": {
                "timestamp": {"system": Decimal("1640611164.396007"), "monotonic": Decimal("4765.594161")},
                "priority": "nominal",
                "transfer_id": 28,
                "source_node_id": 21,
            },
            "timestamp": {"microsecond": 309697890},
            "value": {
                "kinematics": {
                    "angular_position": {"radian": nan},
                    "angular_velocity": {"radian_per_second": 0.0},
                    "angular_acceleration": {"radian_per_second_per_second": 0.0},
                },
                "torque": {"newton_meter": nan},
            },
        }
    }
    tsvfc_formatter = _FORMATTERS["TSVFC"](default_hints)
    assert (
        tsvfc_formatter(obj) == "142{	142._meta_{	142._meta_.timestamp{"
        "	142._meta_.timestamp.system	142._meta_.timestamp.monotonic"
        "	142._meta_.timestamp}	142._meta_.priority	142._meta_.transfer_id"
        "	142._meta_.source_node_id	142._meta_}	142.timestamp{	142.timestamp.microsecond"
        "	142.timestamp}	142.value{	142.value.kinematics{	142.value.kinematics.angular_position{"
        "	142.value.kinematics.angular_position.radian	142.value.kinematics.angular_position}"
        "	142.value.kinematics.angular_velocity{	142.value.kinematics.angular_velocity.radian_per_second"
        "	142.value.kinematics.angular_velocity}	142.value.kinematics.angular_acceleration{"
        "	142.value.kinematics.angular_acceleration.radian_per_second_per_second"
        "	142.value.kinematics.angular_acceleration}	142.value.kinematics}	142.value.torque{"
        "	142.value.torque.newton_meter	142.value.torque}	142.value}	142}\n{	{	{"
        "	1640611164.396007	4765.594161	}	nominal	28	21	}"
        "	{	309697890	}	{	{	{	nan	}	{	0.0	}"
        "	{	0.0	}	}	{	nan	}	}	}\n"
    )
    assert (
        _FORMATTERS["TSV"](default_hints)(obj)
        == "1640611164.396007\t4765.594161\tnominal\t28\t21\t309697890\tnan\t0.0\t0.0\tnan\n"
    )

    assert (
        _FORMATTERS["TSVH"](default_hints)(obj)
        == "142._meta_.timestamp.system\t142._meta_.timestamp.monotonic\t142._meta_.priority\t"
        "142._meta_.transfer_id\t142._meta_.source_node_id\t142.timestamp.microsecond\t"
        "142.value.kinematics.angular_position.radian"
        "\t142.value.kinematics.angular_velocity.radian_per_second\t142.value.kinematics.angular_acceleration."
        "radian_per_second_per_second\t142.value.torque.newton_meter"
        "\n1640611164.396007\t4765.594161\tnominal\t2"
        "8\t21\t309697890\tnan\t0.0\t0.0\tnan\n"
    )



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
from typing import Callable, Any
import click


ProgressCallback = Callable[[str], None]


class ProgressReporter:
    def __init__(self) -> None:
        self._widest = 0
        self._impl = _mk_impl()

    def __call__(self, text: str) -> None:
        self._widest = max(self._widest, len(text))
        # Add extra space after the text is to improve appearance when the text is shortened.
        self._impl(text.ljust(self._widest))

    def clear(self) -> None:
        """
        Call this once at the end to erase the progress line from the screen.
        Does nothing if no output was generated.
        """
        if self._widest > 0:
            self._impl(" " * self._widest)

    def __enter__(self) -> ProgressReporter:
        return self

    def __exit__(self, *_: Any) -> None:
        """
        Invokes :meth:`clear` upon leaving the context.
        """
        self.clear()


def _mk_impl() -> ProgressCallback:
    if sys.stderr.isatty():
        return lambda text: click.secho(f"\r{text}\r", nl=False, file=sys.stderr, fg="green")
    return lambda _: None


def show_error(msg: str) -> None:
    click.secho(msg, err=True, fg="red", bold=True)


def show_warning(msg: str) -> None:
    click.secho(msg, err=True, fg="yellow")



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
from typing import Iterable, Callable, Any, Optional, TYPE_CHECKING
import time
import asyncio
import pycyphal
import yakut

if TYPE_CHECKING:
    import pycyphal.application  # pylint: disable=ungrouped-imports
    from ._controller import ControllerReader, Sample
    from yakut.yaml import EvaluableLoader  # pylint: disable=ungrouped-imports


_logger = yakut.get_logger(__name__)


class Executor:
    SYM_INDEX = "n"
    SYM_TIME = "t"
    SYM_CTRL_AXIS = "A"
    SYM_CTRL_BUTTON = "B"
    SYM_CTRL_TOGGLE = "T"
    SYM_DTYPE = "dtype"

    def __init__(
        self,
        loader: EvaluableLoader,
        publications: Iterable[Publication],
    ) -> None:
        self._ctl: Optional[ControllerReader] = None
        self._publications = list(publications)

        self._loader = loader
        self._loader.evaluation_context[Executor.SYM_CTRL_AXIS] = lambda s, i: self._sample_controller(s).axis[i]
        self._loader.evaluation_context[Executor.SYM_CTRL_BUTTON] = lambda s, i: self._sample_controller(s).button[i]
        self._loader.evaluation_context[Executor.SYM_CTRL_TOGGLE] = lambda s, i: self._sample_controller(s).toggle[i]

    async def run(self, count: int, period: float) -> None:
        started_at: Optional[float] = None
        for index in range(count):
            # Update the expression states. Notice that the controls are sampled once atomically.
            self._loader.evaluation_context[Executor.SYM_INDEX] = index
            self._loader.evaluation_context[Executor.SYM_TIME] = period * index
            if self._ctl:
                self._ctl.sample_and_hold()

            # Compute new messages. The first cycle may be slow to compute due to lazy initialization.
            for pub in self._publications:
                pub.construct_next_message()

            # Run the publication. We initialize the time late to ensure that lazy init doesn't cause phase distortion.
            if started_at is None:
                started_at = asyncio.get_event_loop().time()
            # We must wait for all tasks to complete before raising the exception.
            # Leaving orphans is undesirable as it leads to error handling issues.
            out = await asyncio.gather(*[p.publish() for p in self._publications], return_exceptions=True)
            for ex in out:
                if isinstance(ex, BaseException):
                    raise ex
            assert len(out) == len(self._publications) and all(isinstance(x, bool) for x in out)
            if not all(out):
                timed_out = [self._publications[idx] for idx, res in enumerate(out) if not res]
                _logger.error("The following publications have timed out:\n%s", "\n".join(map(str, timed_out)))

            # Sleep until the next publication cycle.
            sleep_duration = (started_at + period * (index + 1)) - asyncio.get_event_loop().time()
            _logger.info("Published group %6d of %6d; sleeping for %.3f seconds", index + 1, count, sleep_duration)
            await asyncio.sleep(sleep_duration)

    def close(self) -> None:
        pycyphal.util.broadcast(p.close for p in self._publications)()
        if self._ctl:
            self._ctl.close()

    def _sample_controller(self, selector: Any) -> Sample:
        if not self._ctl:
            from ._controller import ControllerReader

            _logger.debug("Constructing the controller reader...")
            started_at = time.monotonic()
            self._ctl = ControllerReader()
            _logger.debug("Done in %.3f sec: %s", time.monotonic() - started_at, self._ctl)
        return self._ctl.read(selector)


class Publication:
    def __init__(
        self,
        subject_id: int,
        dtype: Any,
        evaluator: Callable[..., Any],
        node: "pycyphal.application.Node",
        priority: pycyphal.transport.Priority,
        send_timeout: float,
    ):
        self._dtype = dtype
        self._evaluator = evaluator
        self._publisher = node.make_publisher(self._dtype, subject_id)  # This will expose the port info via registers
        self._publisher.priority = priority
        self._publisher.send_timeout = send_timeout
        self._next_message: Optional[Any] = None
        self._evaluation_context = {
            Executor.SYM_DTYPE: self._dtype,
        }

    def construct_next_message(self) -> None:
        """
        Message construction may take a considerable amount of time depending on the complexity of the expressions.
        This is why they it has to be done separately to keep the publication phase more accurate.
        """
        started_at = time.monotonic()
        # We could make the evaluated expression call a specific function to conditionally cancel publication.
        content = self._evaluator(**self._evaluation_context)
        self._next_message = pycyphal.dsdl.update_from_builtin(self._dtype(), content if content is not None else {})
        _logger.info(
            "%s: Next message (constructed in %.3f sec) shown on the next line:\n%s",
            self,
            time.monotonic() - started_at,
            self._next_message,
        )

    async def publish(self) -> bool:
        if self._next_message is not None:
            msg, self._next_message = self._next_message, None
            out = await self._publisher.publish(msg)
            assert isinstance(out, bool)
            return out
        return True

    def close(self) -> None:
        self._publisher.close()

    def __repr__(self) -> str:
        out = pycyphal.util.repr_attributes(self, self._publisher)
        assert isinstance(out, str)
        return out



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import re
import logging

_logger = logging.getLogger(__name__)


class IntSetError(ValueError):
    pass


INT_SET_USER_DOC = """
Integer set notation examples:

\b
    Discrete elements (, or ;):   1,56;-3
    Intervals [lo,hi) (- or ...): 10-23,-5--7,-10..-2
    Exclusion with ! prefix:      5-9,!6,!5...7
    Arbitrary combination:        -9--5;+4,!-8..-5
    JSON/YAML compatibility:      [1,53,78]
""".strip()


def parse_int_set(text: str) -> set[int] | int:
    """
    Unpacks the integer set notation.
    Accepts JSON-list (subset of YAML) of integers at input, too.
    A single scalar is returned as-is unless there is a separator at the end ("125,") or JSON list is used.
    Raises :class:`IntSetError` on syntax error.
    Usage:

    >>> parse_int_set("")
    set()
    >>> parse_int_set("123"), parse_int_set("[123]"), parse_int_set("123,")
    (123, {123}, {123})
    >>> parse_int_set("-0"), parse_int_set("[-0]"), parse_int_set("-0,")
    (0, {0}, {0})
    >>> sorted(parse_int_set("0..0x0A"))    # Half-open interval with .. or ... or -
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> sorted(parse_int_set("-9...-5,"))
    [-9, -8, -7, -6]
    >>> sorted(parse_int_set("-9--5; +4, !-8..-5"))     # Exclusion with ! prefix
    [-9, 4]
    >>> sorted(parse_int_set("-10..+10,!-9-+9"))    # Valid separators are , and ;
    [-10, 9]
    >>> sorted(parse_int_set("6-6"))
    []
    >>> sorted(parse_int_set("[1,53,78]"))
    [1, 53, 78]
    >>> parse_int_set("123,456,9-") # doctest:+IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    IntSetError: ...
    """

    def try_parse(val: str) -> int | None:
        try:
            return int(val, 0)
        except ValueError:
            return None

    collapse = not _RE_JSON_LIST.match(text)
    incl: set[int] = set()
    excl: set[int] = set()
    for item in _RE_SPLIT.split(_RE_JSON_LIST.sub(r"\1", text)):
        item = item.strip()
        if not item:
            collapse = False
            continue
        if item.startswith("!"):
            target_set = excl
            item = item[1:]
        else:
            target_set = incl
        x = try_parse(item)
        if x is not None:
            target_set.add(x)
            continue
        match = _RE_RANGE.match(item)
        if match:
            lo, hi = map(try_parse, match.groups())
            if lo is not None and hi is not None:
                target_set |= set(range(lo, hi))
                continue
        raise IntSetError(f"Item {item!r} of the integer set {text!r} could not be parsed")

    result: set[int] | int = incl - excl
    assert isinstance(result, set)
    if collapse and len(result) == 1:
        (result,) = result
    _logger.debug("Int set %r parsed as %r", text, result)
    return result


_RE_JSON_LIST = re.compile(r"^\s*\[([^]]*)]\s*$")
_RE_SPLIT = re.compile(r"[,;]")
_RE_RANGE = re.compile(r"([+-]?\w+)(?:-|\.\.\.?)([+-]?\w+)")



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
from typing import Iterable, Tuple, Callable, Dict, Optional, List
import sys
import functools
import threading
import yakut
from . import Controller, Sample, ControllerError, ControllerNotFoundError

try:
    import sdl2  # type: ignore
except ImportError as _ex:
    raise ImportError(
        "Joystick support not installed or libsdl2 is missing. "
        "Please install optional dependencies with yakut[joystick] "
        "and check https://libsdl.org"
    ) from _ex


class JoystickController(Controller):
    """
    Interface for standard HID joystick.
    """

    def __init__(self, index: int) -> None:
        self._handle = sdl2.joystick.SDL_JoystickOpen(index)
        if not self._handle:
            raise ControllerNotFoundError(f"Cannot open joystick {index}")
        self._id = sdl2.joystick.SDL_JoystickInstanceID(self._handle)
        self._name = str(sdl2.joystick.SDL_JoystickNameForIndex(index).decode())
        self._update_hook: Callable[[], None] = lambda: None

        n_axes = sdl2.joystick.SDL_JoystickNumAxes(self._handle)
        n_hats = sdl2.joystick.SDL_JoystickNumHats(self._handle)
        n_buttons = sdl2.joystick.SDL_JoystickNumButtons(self._handle)

        self._axes: List[float] = [
            JoystickController._scale_axis(sdl2.joystick.SDL_JoystickGetAxis(self._handle, i)) for i in range(n_axes)
        ]
        self._hats: List[Tuple[int, int]] = [
            JoystickController._split_hat(sdl2.joystick.SDL_JoystickGetHat(self._handle, i)) for i in range(n_hats)
        ]
        self._buttons: List[bool] = [sdl2.joystick.SDL_JoystickGetButton(self._handle, i) for i in range(n_buttons)]
        self._counters: List[int] = [0 for _ in self._buttons]

        _registry[self._id] = self._callback

        _logger.info(
            "%s: Joystick %r initial state: axes=%s hats=%s buttons=%s",
            self,
            index,
            self._axes,
            self._hats,
            self._buttons,
        )

    @property
    def name(self) -> str:
        return self._name

    def sample(self) -> Sample:
        with _lock:
            if _exception:
                raise ControllerError("Worker thread failed") from _exception

            axes_and_hats = self._axes.copy()
            for x, y in self._hats:
                axes_and_hats.append(float(x))
                axes_and_hats.append(float(y))

            return Sample(
                axis=dict(enumerate(axes_and_hats)),
                button=dict(enumerate(self._buttons)),
                toggle={k: v % 2 != 0 for k, v in enumerate(self._counters)},
            )

    def set_update_hook(self, hook: Callable[[], None]) -> None:
        self._update_hook = hook

    def close(self) -> None:
        with _lock:
            _registry.pop(self._id, None)
            sdl2.joystick.SDL_JoystickClose(self._handle)

    def _callback(self, event: sdl2.SDL_Event) -> None:
        if event.type == sdl2.SDL_JOYAXISMOTION:
            self._axes[event.jaxis.axis] = JoystickController._scale_axis(event.jaxis.value)

        elif event.type in (sdl2.SDL_JOYBUTTONDOWN, sdl2.SDL_JOYBUTTONUP):
            if event.jbutton.state == sdl2.SDL_PRESSED:
                self._buttons[event.jbutton.button] = True
                self._counters[event.jbutton.button] += 1
            else:
                self._buttons[event.jbutton.button] = False

        elif event.type == sdl2.SDL_JOYHATMOTION:
            self._hats[event.jhat.hat] = JoystickController._split_hat(event.jhat.value)

        else:
            _logger.debug("%s: Event dropped: %r", self, event)

        self._update_hook()

    @staticmethod
    def _scale_axis(raw: int) -> float:
        if raw >= 0:
            return raw / 32767.0
        return raw / 32768.0

    @staticmethod
    def _split_hat(value: int) -> Tuple[int, int]:
        return (
            (bool(value & sdl2.SDL_HAT_RIGHT) - bool(value & sdl2.SDL_HAT_LEFT)),
            (bool(value & sdl2.SDL_HAT_UP) - bool(value & sdl2.SDL_HAT_DOWN)),
        )

    @staticmethod
    def list_controllers() -> Iterable[Tuple[str, Callable[[], Controller]]]:
        def construct(index: int) -> Controller:
            with _lock:
                return JoystickController(index)

        if not _init_done.wait(10.0):  # pragma: no cover
            raise _exception or ControllerError("The worker thread has failed to initialize")

        with _lock:
            num_joys = sdl2.joystick.SDL_NumJoysticks()
            for idx in range(num_joys):
                name = sdl2.joystick.SDL_JoystickNameForIndex(idx).decode()
                yield name, functools.partial(construct, idx)


_exception: Optional[Exception] = None
_lock = threading.RLock()
_init_done = threading.Event()
_registry: Dict[sdl2.SDL_JoystickID, Callable[[sdl2.SDL_Event], None]] = {}


def _dispatch_joy(joystick: sdl2.SDL_JoystickID, event: sdl2.SDL_Event) -> None:
    with _lock:
        try:
            _registry[joystick](event)
        except KeyError:
            _logger.debug("No handler for joystick %r; dropping event %r", joystick, event)


def _run_sdl2() -> None:
    # Shall we require SDL2 somewhere else in this app, this logic will have to be extracted into a shared component.
    global _exception  # pylint: disable=global-statement
    try:
        import ctypes

        # Initialization and event processing should be done in the same thread.
        init_subsystems = sdl2.SDL_INIT_JOYSTICK
        if sys.platform.startswith("win"):  # pragma: no cover
            # I don't understand why is this necessary but joystick does not work on Windows without the video subsystem
            init_subsystems |= sdl2.SDL_INIT_VIDEO
        err = sdl2.SDL_Init(init_subsystems)
        if err != 0:
            raise ControllerError(f"Could not initialize SDL2: {sdl2.SDL_GetError()!r}")

        sdl2.SDL_JoystickEventState(sdl2.SDL_ENABLE)
        sdl2.SDL_SetHint(sdl2.SDL_HINT_JOYSTICK_ALLOW_BACKGROUND_EVENTS, b"1")

        _logger.debug("SDL2 initialized successfully, entering the event loop")
        _init_done.set()

        event = sdl2.SDL_Event()
        while True:
            if sdl2.SDL_WaitEvent(ctypes.byref(event)) != 1:
                raise ControllerError(f"Could not poll event: {sdl2.SDL_GetError()!r}")

            if event.type == sdl2.SDL_JOYAXISMOTION:
                _dispatch_joy(event.jaxis.which, event)
            elif event.type == sdl2.SDL_JOYBALLMOTION:
                _dispatch_joy(event.jball.which, event)
            elif event.type in (sdl2.SDL_JOYBUTTONDOWN, sdl2.SDL_JOYBUTTONUP):
                _dispatch_joy(event.jbutton.which, event)
            elif event.type == sdl2.SDL_JOYHATMOTION:
                _dispatch_joy(event.jhat.which, event)
            else:
                _logger.debug("Event dropped: %r", event)

    except Exception as ex:  # pylint: disable=broad-except
        _exception = ex
        _logger.exception("SDL2 worker thread failed: %s", ex)


_logger = yakut.get_logger(__name__)

threading.Thread(target=_run_sdl2, name="sdl2_worker", daemon=True).start()



# Copyright (c) 2020 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import os
import sys
import asyncio
import functools
from typing import TYPE_CHECKING, Iterable, Optional, Any, Callable, Awaitable
import logging
from pathlib import Path
from shutil import get_terminal_size
import click
import yakut
from yakut.param.transport import transport_factory_option, TransportFactory, Transport
from yakut.param.formatter import formatter_factory_option, FormatterFactory, Formatter, FormatterHints
from yakut.param.node import node_factory_option, NodeFactory

if TYPE_CHECKING:
    import pycyphal.application


def get_logger(name: str) -> logging.Logger:
    """
    This is a trivial wrapper over :func:`logging.getLogger` that removes private components from the logger name.
    For example, ``yakut.cmd.file_server._cmd`` becomes ``yakut.cmd.file_server`` (private submodule hidden).
    Also, double underscores are removed.
    All this is done to make the log messages appear nicer, since this is important for a CLI tool.
    """
    return logging.getLogger(name.replace("__", "").split("._", 1)[0])


_logger = get_logger("yakut")
# Some of the integration tests may parse the logs expecting its lines to follow this format.
# If you change this, you may break these tests.
_LOG_FORMAT = "%(asctime)s %(process)07d %(levelname)-3.3s %(name)s: %(message)s"
logging.basicConfig(format=_LOG_FORMAT)  # Using the default log level; it will be overridden later.


class Purser:
    def __init__(
        self,
        paths: Iterable[str | Path],
        formatter_factory: FormatterFactory,
        transport_factory: TransportFactory,
        node_factory: NodeFactory,
    ) -> None:
        self._paths = list(Path(x) for x in paths)
        self._f_formatter = formatter_factory
        self._f_transport = transport_factory
        self._f_node = node_factory

        self._registry: Optional[pycyphal.application.register.Registry] = None
        self._transport: Optional[Transport] = None
        self._node: Optional["pycyphal.application.Node"] = None

    @property
    def paths(self) -> list[Path]:
        return list(self._paths)

    def make_formatter(self, hints: FormatterHints = FormatterHints()) -> Formatter:
        return self._f_formatter(hints)

    def get_registry(self) -> pycyphal.application.register.Registry:
        """
        Commands should never construct registry on their own!
        Doing so is likely to create divergent configurations that are not exposed via the Register Interface.
        Instead, use this factory: it will create a registry instance at the first invocation and then it will be
        shared with all components.

        :raises: :class:`ImportError` if the standard DSDL namespace ``uavcan`` is not available.
        """
        if self._registry is None:
            from pycyphal.application import make_registry

            self._registry = make_registry()
        return self._registry

    def get_transport(self) -> Transport:
        if self._transport is None:  # pragma: no branch
            self._transport = self._f_transport()
        if self._transport is not None:
            return self._transport
        click.get_current_context().fail("Transport not configured, or the standard DSDL namespace is not compiled")

    def get_node(self, name_suffix: str, allow_anonymous: bool) -> "pycyphal.application.Node":
        if self._node is None:  # pragma: no branch
            tr = self.get_transport()
            self._node = self._f_node(tr, name_suffix=name_suffix, allow_anonymous=allow_anonymous)
        return self._node


pass_purser = click.make_pass_decorator(Purser)


class AliasedGroup(click.Group):
    """
    This class is inspired by "click-aliases" from Robbin Bonthond published at
    https://github.com/click-contrib/click-aliases/blob/master/click_aliases/__init__.py
    under the terms of the MIT license.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._commands: dict[str, set[str]] = {}
        self._aliases: dict[str, str] = {}

    def command(self, *args: Any, **kwargs: Any) -> Any:
        aliases = AliasedGroup._mk_aliases(kwargs.pop("aliases", []))
        decorator: Any = super().command(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f: Any) -> Any:
            cmd: Any = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def group(self, *args: Any, **kwargs: Any) -> Any:
        aliases = AliasedGroup._mk_aliases(kwargs.pop("aliases", []))
        decorator: Any = super().group(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f: Any) -> Any:
            cmd: Any = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def get_command(self, ctx: click.Context, cmd_name: str) -> Any:
        cmd_name = self._aliases.get(cmd_name, cmd_name)
        return super().get_command(ctx, cmd_name)

    def resolve_command(
        self, ctx: click.Context, args: list[Any]
    ) -> tuple[str | None, click.Command | None, list[str]]:
        """
        This is a workaround for this bug in v7 and v8: https://github.com/pallets/click/issues/1422.
        If this is not overridden, then abbreviated commands cause the automatic envvar prefix to be constructed
        incorrectly, such that instead of the full command name the abbreviated name is used.
        For example, if the user invokes `yakut co` meaning `yakut compile`,
        the auto-constructed envvar prefix would be `YAKUT_CO_` instead of `YAKUT_COMPILE_`.
        """
        _, cmd, out_args = super().resolve_command(ctx, args)
        return (cmd.name if cmd else None), cmd, out_args

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        rows: list[tuple[str, str]] = []
        sub_commands = self.list_commands(ctx)
        max_len = max(len(cmd) for cmd in sub_commands)
        limit = formatter.width - 6 - max_len
        for subcmd in sub_commands:
            cmd = self.get_command(ctx, subcmd)
            if cmd is not None and not getattr(cmd, "hidden", False):
                if subcmd in self._commands:
                    subcmd = ",".join([subcmd] + list(sorted(self._commands[subcmd])))
                rows.append((subcmd, cmd.get_short_help_str(limit)))
        if rows:
            with formatter.section("Commands (with aliases)"):
                formatter.write_dl(rows)

    @staticmethod
    def _mk_aliases(item: Any) -> set[str]:
        if isinstance(item, str):
            return {item}
        if isinstance(item, (list, tuple, set)) and all(isinstance(x, str) for x in item):
            return set(item)
        raise TypeError(f"Bad aliases: {item}")


_ENV_VAR_PATH = "YAKUT_PATH"


@click.command(
    cls=AliasedGroup,
    context_settings={
        "max_content_width": get_terminal_size()[0],
        "auto_envvar_prefix": "YAKUT",  # Specified here, not in __main__.py, otherwise doesn't work when installed.
    },
)
@click.version_option(version=yakut.__version__)
@click.option("--verbose", "-v", count=True, help="Emit verbose log messages. Specify twice for extra verbosity.")
@click.option(
    "--path",
    "-P",
    multiple=True,
    type=click.Path(resolve_path=True),
    help=f"""
In order to use compiled DSDL namespaces,
the directories that contain compilation outputs need to be specified using this option.
The current working directory does not need to be specified explicitly.

Examples:

\b
    yakut  --path ../public_regulated_data_types  --path ~/my_namespaces  pub ...

\b
    export {_ENV_VAR_PATH}="../public_regulated_data_types:~/my_namespaces"
    yakut pub ...
""",
)
@formatter_factory_option
@transport_factory_option
@node_factory_option
@click.pass_context
def _click_main(
    ctx: click.Context,
    verbose: int,
    path: tuple[str, ...],
    formatter_factory: FormatterFactory,
    transport_factory: TransportFactory,
    node_factory: NodeFactory,
) -> None:
    """
    \b
           ____                   ______            __          __
          / __ `____  ___  ____  / ____/_  ______  / /_  ____  / /
         / / / / __ `/ _ `/ __ `/ /   / / / / __ `/ __ `/ __ `/ /
        / /_/ / /_/ /  __/ / / / /___/ /_/ / /_/ / / / / /_/ / /
        `____/ .___/`___/_/ /_/`____/`__, / .___/_/ /_/`__,_/_/
            /_/                     /____/_/

    Yakut is a cross-platform command-line utility for diagnostics and management of Cyphal networks.
    It is designed for use either directly by humans or from automation scripts.
    Ask questions at https://forum.opencyphal.org

    Any long option can be provided via environment variable prefixed with `YAKUT_`
    such that an option `--foo-bar` for command `baz`, if not provided as a command-line argument,
    will be read from `YAKUT_BAZ_FOO_BAR`.
    """
    _configure_logging(verbose)  # This should be done in the first order to ensure that we log things correctly.

    path = (os.getcwd(), *path)
    _logger.debug("Path: %r", path)
    for p in path:
        sys.path.append(str(p))

    ctx.obj = Purser(
        paths=path,
        formatter_factory=formatter_factory,
        transport_factory=transport_factory,
        node_factory=node_factory,
    )


def main() -> None:  # https://click.palletsprojects.com/en/8.1.x/exceptions/
    from yakut.ui import show_error

    status: Any = 1
    # noinspection PyBroadException
    try:
        status = _click_main.main(prog_name="yakut", standalone_mode=False)

    except SystemExit as ex:
        status = ex.code

    except (KeyboardInterrupt, click.Abort) as ex:
        status = 127
        _logger.info("Interrupted")
        _logger.debug("%s: %s", type(ex).__name__, ex, exc_info=True)

    except click.ClickException as ex:
        status = ex.exit_code
        try:
            click.secho("", err=True, fg="red", bold=True, reset=False, nl=False)
            ex.show()
        finally:
            click.secho("", err=True, nl=False)

    except Exception as ex:  # pylint: disable=broad-except
        show_error(f"{type(ex).__name__}: {ex}")
        _logger.debug("EXCEPTION %s: %s", type(ex).__name__, ex, exc_info=True)

    except BaseException as ex:  # pylint: disable=broad-except
        show_error(f"Internal error, please report: {ex}")
        _logger.error("%s: %s", type(ex).__name__, ex, exc_info=True)

    _logger.debug("EXIT %r", status)
    sys.exit(status)


subcommand: Callable[..., Callable[..., Any]] = _click_main.command  # type: ignore


def asynchronous(*, interrupted_ok: bool = False) -> Callable[[Callable[..., Awaitable[Any]]], Callable[..., Any]]:
    def impl(f: Callable[..., Awaitable[Any]]) -> Callable[..., Any]:
        def handle_task_exception(_loop: asyncio.AbstractEventLoop, context: dict[str, Any]) -> None:
            message = context.get("message", "Unhandled exception in event loop")
            exc = context.get("exception")
            exc_info: Any = (type(exc), exc, exc.__traceback__) if exc else False
            _logger.debug("Task exception during shutdown: %s", message, exc_info=exc_info)

        # This is similar to asyncio.run().
        # The difference is that we configure logging differently to avoid unhelpful stack traces during shutdown.
        # Appearance of such errors is not the expected behavior and should be fixed, but the user need not
        # know about these errors as they are unlikely to affect the behavior of the application.
        # See https://github.com/OpenCyphal/yakut/issues/40
        def proxy(*args: Any, **kwargs: Any) -> Any:
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                return loop.run_until_complete(f(*args, **kwargs))
            except KeyboardInterrupt:
                if not interrupted_ok:
                    raise
            finally:
                _logger.debug("Event loop finalization with exc=%r", sys.exc_info())
                try:
                    loop.set_exception_handler(handle_task_exception)  # Reduce severity of exception reports

                    # Suppress finalization errors from PyCyphal https://github.com/OpenCyphal/pycyphal/issues/227
                    pycyphal_logger = logging.getLogger("pycyphal")
                    if not pycyphal_logger.isEnabledFor(logging.INFO):  # Do not suppress if verbose
                        pycyphal_logger.setLevel(logging.CRITICAL)

                    orphans = asyncio.all_tasks(loop)
                    if orphans:
                        for ts in orphans:
                            ts.cancel()
                        for e in loop.run_until_complete(asyncio.gather(*orphans, return_exceptions=True)):
                            if isinstance(e, BaseException) and not isinstance(e, asyncio.CancelledError):
                                handle_task_exception(loop, {"exception": e})
                finally:
                    asyncio.set_event_loop(None)
                    loop.close()

        return functools.update_wrapper(proxy, f)

    return impl


def _configure_logging(verbosity_level: int) -> None:
    log_level = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }.get(verbosity_level or 0, logging.DEBUG)

    logging.root.setLevel(log_level)

    try:
        import coloredlogs  # type: ignore

        # The level spec applies to the handler, not the root logger! This is different from basicConfig().
        coloredlogs.install(level=log_level, fmt=_LOG_FORMAT)
    except Exception as ex:  # pylint: disable=broad-except
        _logger.exception("Could not set up coloredlogs: %r", ex)  # pragma: no cover



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

# pylint: disable=too-many-boolean-expressions,too-many-instance-attributes

from __future__ import annotations
import re
from typing import Optional, TYPE_CHECKING, Any
import dataclasses
import pycyphal
import yakut

if TYPE_CHECKING:
    import pycyphal.application  # pylint: disable=ungrouped-imports


_logger = yakut.get_logger(__name__.split("._", maxsplit=1)[0])


@dataclasses.dataclass(frozen=True)
class AppDescriptor:
    r"""
    Application package file name format::

        NAME-HW_MAJ.HW_MIN-SW_MAJ.SW_MIN.SW_VCS.SW_CRC.app*
                   \_____/                     \_____/
            \____________/              \____________/
          either major or both          either CRC or both
        can be omitted if multiple      CRC and VCS can be
       hardware versions supported      omitted if irrelevant

    The values are sourced from uavcan.node.GetInfo, and they are as follows:

    NAME -- The name of the node; e.g., "com.zubax.telega".

    HW_MAJ, HW_MIN --
    Hardware version numbers.
    The minor number or both of them can be omitted iff the package is compatible with multiple hardware revisions.

    SW_MAJ, SW_MIN -- Software version numbers.

    SW_VCS, SW_CRC --
    The version control system (VCS) revision ID (e.g., git commit hash) and the CRC of the software package.
    Both are hexadecimal numbers and both are optional: either the CRC alone or both VCS-hash and CRC can be omitted
    (CRC cannot be populated without the VCS-hash).

    The fields are terminated by a literal string ".app",
    which can be followed by arbitrary additional metadata (like a file extension).

    Examples: com.zubax.telega-1.2-0.3.68620b82.28df0c432c2718cd.app.bin, com.zubax.telega-0.3.app

    The implementation applies trivial heuristics to decide if a given application
    needs to be updated to another version.
    Generally, an update is considered necessary if the available hardware version numbers match and
    CRC is different (if defined) or the update is of a newer version.
    """

    FILE_NAME_PATTERN = re.compile(
        r"(?i)([\w.]+)(?:-(\d+)(?:\.(\d+))?)?-(\d+)\.(\d+)(?:\.([0-9a-f]+)(?:\.([0-9a-f]+))?)?\.app.*"
    )

    name: str
    hw_maj: Optional[int]
    hw_min: Optional[int]
    sw_maj: int
    sw_min: int
    sw_vcs: Optional[int]
    sw_crc: Optional[int]

    def is_equivalent(self, other: AppDescriptor) -> bool:
        """
        :return: True if updating "self" to "other" is pointless because they are likely to be functionally equivalent.
            A property that is not defined for at least one operand is considered to match.
            For instance, if CRC is not defined, it is considered to match.
        """

        def compatible(a: Optional[int], b: Optional[int]) -> bool:
            if a is None or b is None:
                return True
            return a == b

        self._log("Checking if %s is equivalent...", other)
        if self.name != other.name:
            self._log("No, it is a completely different application.")
            return False
        if not compatible(self.hw_maj, other.hw_maj) or not compatible(self.hw_min, other.hw_min):
            self._log("No, the hardware version does not match.")
            return False
        if self.sw_maj != other.sw_maj or self.sw_min != other.sw_min:
            self._log("No, the software version does not match.")
            return False
        if not compatible(self.sw_vcs, other.sw_vcs) or not compatible(self.sw_crc, other.sw_crc):
            self._log("No, VCS/CRC do not match.")
            return False
        self._log("Yes, this app is equivalent to %s, an update is likely to be meaningless.", other)
        return True

    def should_update_to(self, other: AppDescriptor) -> bool:
        self._log("Checking if this app should be updated to %s...", other)
        if self.name != other.name:
            self._log("No, this is a completely different application.")
            return False
        if (self.hw_maj is not None and other.hw_maj is not None and self.hw_maj != other.hw_maj) or (
            self.hw_min is not None and other.hw_min is not None and self.hw_min != other.hw_min
        ):
            self._log("No, the hardware version does not match, an update might brick the node.")
            return False

        self._log("The applications appear to be compatible -- an update would not break the node. Looking further...")
        if self.sw_crc is not None and other.sw_crc is not None and self.sw_crc != other.sw_crc:
            self._log("The CRC is different! An update is therefore required regardless of the other parameters.")
            return True

        self._log("CRC is identical or not defined, checking the version information...")
        if (self.sw_maj > other.sw_maj) or ((self.sw_maj == other.sw_maj) and (self.sw_min > other.sw_min)):
            self._log("No, the other application is older.")
            return False

        vcs_different = self.sw_vcs is not None and other.sw_vcs is not None and self.sw_vcs != other.sw_vcs
        if (self.sw_maj != other.sw_maj) or (self.sw_min != other.sw_min) or vcs_different:
            self._log("Yes, %s is of a different version and is not older than the current one.", other)
            return True

        self._log("No, this application does not need to be updated to %s", other)
        return False

    def make_glob_expression(self) -> str:
        """
        Construct a glob expression for pre-matching.
        This is helpful when you need to filter out relevant files from a large directory.
        """
        return f"{self.name}-*.app*"

    @staticmethod
    def from_file_name(file_name: str) -> Optional[AppDescriptor]:
        match = AppDescriptor.FILE_NAME_PATTERN.match(file_name)
        if not match:
            return None
        name, hw_maj, hw_min, sw_maj, sw_min, sw_vcs, sw_crc = match.groups()

        def mint10(s: str) -> Optional[int]:
            return int(s, 10) if s is not None else None

        def mint16(s: str) -> Optional[int]:
            return int(s, 16) if s is not None else None

        return AppDescriptor(
            name.lower(),
            mint10(hw_maj),
            mint10(hw_min),
            int(sw_maj),
            int(sw_min),
            mint16(sw_vcs),
            mint16(sw_crc),
        )

    @staticmethod
    def from_node_info(info: pycyphal.application.NodeInfo) -> AppDescriptor:
        has_hw = info.hardware_version.major > 0 or info.hardware_version.minor > 0
        return AppDescriptor(
            info.name.tobytes().decode(errors="ignore").strip().lower(),
            info.hardware_version.major if has_hw else None,
            info.hardware_version.minor if has_hw else None,
            info.software_version.major,
            info.software_version.minor,
            info.software_vcs_revision_id or None,
            info.software_image_crc.sum() or None,
        )

    def _log(self, text: str, *args: Any) -> None:
        text = f"{self}: {text}"
        _logger.info(text, *args)

    def __str__(self) -> str:
        out = f"{self.name}"
        if self.hw_maj is not None:
            out += f"-{self.hw_maj}" + f".{self.hw_min}" * (self.hw_min is not None)
        out += f"-{self.sw_maj}.{self.sw_min}"
        if self.sw_vcs is not None:
            out += f".{self.sw_vcs:016x}" + (f".{self.sw_crc:016x}" if self.sw_crc is not None else "")
        return out + ".app"


def _unittest_app_descriptor_from_node_info() -> None:
    from tests.dsdl import ensure_compiled_dsdl

    ensure_compiled_dsdl()

    from pycyphal.application import NodeInfo
    from uavcan.node import Version_1_0 as Version

    ad = AppDescriptor.from_node_info(
        NodeInfo(
            hardware_version=Version(16, 17),
            software_version=Version(26, 27),
            software_vcs_revision_id=0x123456,
            name="org.opencyphal.NODE",
            software_image_crc=[0xDEADBEEF],
        )
    )
    assert ad.name == "org.opencyphal.node"
    assert ad.hw_maj == 16
    assert ad.hw_min == 17
    assert ad.sw_maj == 26
    assert ad.sw_min == 27
    assert ad.sw_vcs == 0x123456
    assert ad.sw_crc == 0xDEADBEEF

    ad = AppDescriptor.from_node_info(
        NodeInfo(
            software_version=Version(26, 27),
            name="org.opencyphal.NODE",
        )
    )
    assert ad.name == "org.opencyphal.node"
    assert ad.hw_maj is None
    assert ad.hw_min is None
    assert ad.sw_maj == 26
    assert ad.sw_min == 27
    assert ad.sw_vcs is None
    assert ad.sw_crc is None


def _unittest_app_descriptor_from_file_name() -> None:
    assert "org.opencyphal.node-16.17-26.27.0000000000123456.00000000deadbeef.app" == str(
        AppDescriptor.from_file_name("org.opencyphal.NODE-16.17-26.27.123456.DEADBEEF.application.bin")
    )
    assert "org.opencyphal.node-16-26.27.0000000000123456.00000000deadbeef.app" == str(
        AppDescriptor.from_file_name("org.opencyphal.NODE-16-26.27.123456.DEADBEEF.application.bin")
    )
    assert "org.opencyphal.node-26.27.0000000000123456.00000000deadbeef.app" == str(
        AppDescriptor.from_file_name("org.opencyphal.NODE-26.27.123456.DEADBEEF.application.bin")
    )
    assert "org.opencyphal.node-26.27.0000000000123456.app" == str(
        AppDescriptor.from_file_name("org.opencyphal.NODE-26.27.123456.application.bin")
    )
    assert "org.opencyphal.node-26.27.app" == str(AppDescriptor.from_file_name("org.opencyphal.NODE-26.27.app"))
    assert "org.opencyphal.node-16.17-26.27.app" == str(
        AppDescriptor.from_file_name("org.opencyphal.NODE-16.17-26.27.app")
    )

    assert None is AppDescriptor.from_file_name("org.opencyphal.node-z-26.27.app")
    assert None is AppDescriptor.from_file_name("org.opencyphal.NODE-16.17-26.27.123456.DEADBEEF.bin")


def _unittest_app_descriptor_equivalency() -> None:
    def ffn(s: str) -> AppDescriptor:
        x = AppDescriptor.from_file_name(s)
        assert x
        return x

    # Missing parameters on either side are matched as equivalent.
    assert ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.2-3.4.a.b.app"))
    assert ffn("z-1.2-3.4.a.app").is_equivalent(ffn("z-1.2-3.4.a.b.app"))
    assert ffn("z-1.2-3.4.app").is_equivalent(ffn("z-1.2-3.4.a.b.app"))
    assert ffn("z-1-3.4.app").is_equivalent(ffn("z-1.2-3.4.a.b.app"))
    assert ffn("z-3.4.app").is_equivalent(ffn("z-1.2-3.4.a.b.app"))
    assert ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.2-3.4.a.app"))
    assert ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.2-3.4.app"))
    assert ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1-3.4.a.b.app"))
    assert ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-3.4.a.b.app"))

    assert not ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("x-1.2-3.4.a.b.app"))  # Name
    assert not ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-2.2-3.4.a.b.app"))  # Hw major
    assert not ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.3-3.4.a.b.app"))  # Hw minor
    assert not ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.2-4.4.a.b.app"))  # Sw major
    assert not ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.2-3.3.a.b.app"))  # Sw minor
    assert not ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.2-3.4.c.b.app"))  # VCS
    assert not ffn("z-1.2-3.4.a.b.app").is_equivalent(ffn("z-1.2-3.4.a.c.app"))  # CRC


def _unittest_app_descriptor_update() -> None:
    def ffn(s: str) -> AppDescriptor:
        x = AppDescriptor.from_file_name(s)
        assert x
        return x

    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.4.a.b.app"))  # Base, same

    assert ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-4.4.a.b.app"))  # Sw major
    assert ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.5.a.b.app"))  # Sw minor

    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.3.a.b.app"))  # rhs is older
    assert ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.3.a.c.app"))  # CRC is different, version ignored
    assert ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.3.a.c.app"))  # CRC is different, version ignored
    assert ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-3.3.a.c.app"))  # CRC is different, version ignored

    assert ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.4.c.b.app"))  # VCS hash differs
    assert ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.4.c.app"))  # Same without CRC
    assert ffn("z-1.2-3.4.a.app").should_update_to(ffn("z-1.2-3.4.c.b.app"))  # Same without CRC

    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("x-1.2-3.3.a.c.app"))  # Wrong name
    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("x-1.3-3.3.a.c.app"))  # Wrong hardware
    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("x-2.2-3.3.a.c.app"))  # Wrong hardware

    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.4.a.b.app"))
    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-2.4.a.b.app"))  # Sw major
    assert not ffn("z-1.2-3.4.a.b.app").should_update_to(ffn("z-1.2-3.3.a.b.app"))  # Sw minor



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
from typing import Any, Sequence, TYPE_CHECKING, Optional, Callable
import click
import pycyphal
import yakut
from yakut.int_set_parser import parse_int_set, INT_SET_USER_DOC
from yakut.ui import ProgressReporter, show_error, show_warning
from yakut.param.formatter import FormatterHints
from yakut.register import explode_value, get_access_response_metadata
from yakut.util import EXIT_CODE_UNSUCCESSFUL
from ._logic import access

if TYPE_CHECKING:
    import pycyphal.application
    from uavcan.register import Access_1

_logger = yakut.get_logger(__name__)


_HELP = f"""
Read or modify a register on one or multiple remote nodes.

If no value is given, the register will be only read from the specified nodes,
and only one request will be sent per node.

If a value is given, it shall follow the standard environment variable notation described in the
uavcan.register.Access service specification.
The assignment value will be parsed depending on the type of the register reported by the remote node
(which may be different per node).

If a value is given, two requests will be sent: the first one to discover the type,
the second to actually write the value
(unless the provided value could not be applied to the register type).
The final value returned by the node is always printed at the end,
which may be different from the assigned value depending on the internal logic of the node
(e.g., the value could be adjusted to satisfy constraints, the register could be read-only, etc.).

The output (but not the behavior) depends on how the node-ID set is specified:
if it's a single number, the result will be only the value of the specified register;
if multiple node-IDs are given or the sole node-ID is given in the set notation,
the output will be a map of (node_id->register_value).

Examples:

\b
    yakut reg 125  m.inductance_dq                      # Outputs only the value
    yakut reg 125, m.inductance_dq                      # Outputs {{125: value}}
    yakut reg 122-126 m.inductance_dq '12.0e-6 14.7e-6'
    yakut reg 122-126 m.inductance_dq  12.0e-6 14.7e-6  # Quotes are optional
    y r 122-126 m.inductance_dq | jq '[.[]]|unique'     # Remove duplicate values from output
    y r 125 uavcan.node.description "Motor rear-left #3"

{INT_SET_USER_DOC}
"""


@yakut.subcommand(aliases=["r", "reg"], help=_HELP)
@click.argument("node_ids", type=parse_int_set)
@click.argument("register_name")
@click.argument("register_value_element", nargs=-1)
@click.option(
    "--timeout",
    "-T",
    type=float,
    default=pycyphal.presentation.DEFAULT_SERVICE_REQUEST_TIMEOUT,
    show_default=True,
    metavar="SECONDS",
    help="Service response timeout.",
)
@click.option(
    "--optional-service",
    "-s",
    is_flag=True,
    help="""
Ignore nodes that fail to respond to the first RPC-service request instead of reporting an error
assuming that the register service is not supported.
If a node responded at least once it is assumed to support the service and any future timeout
will be always treated as error.
Best-effort output will always be produced regardless of this option; that is, it only affects the exit code.
""",
)
@click.option(
    "--optional-register",
    "-r",
    is_flag=True,
    help="""
If modification is requested (i.e., if a value is given),
nodes that report that they don't have such register are silently ignored instead of reporting an error.
Best-effort output will always be produced regardless of this option; that is, it only affects the exit code.
""",
)
@click.option(
    "--detailed",
    "-d",
    count=True,
    help="Display the value as-is, with DSDL type information, do not simplify. Specify twice to also show metadata.",
)
@yakut.pass_purser
@yakut.asynchronous()
async def register_access(
    purser: yakut.Purser,
    node_ids: Sequence[int] | int,
    register_name: str,
    register_value_element: Sequence[str],
    timeout: float,
    optional_service: bool,
    optional_register: bool,
    detailed: int,
) -> int:
    _logger.debug(
        "node_ids=%r, register_name=%r, register_value_element=%r timeout=%r",
        node_ids,
        register_name,
        register_value_element,
        timeout,
    )
    _logger.debug(
        "optional_service=%r optional_register=%r detailed=%r",
        optional_service,
        optional_register,
        detailed,
    )
    reg_val_str = " ".join(register_value_element) if len(register_value_element) > 0 else None
    del register_value_element
    formatter = purser.make_formatter(FormatterHints(single_document=True))
    representer = _make_representer(simplify=detailed < 1, metadata=detailed > 1)

    with purser.get_node("register_access", allow_anonymous=False) as node:
        with ProgressReporter() as prog:
            result = await access(
                node,
                prog,
                list(sorted(node_ids)) if not isinstance(node_ids, int) else [node_ids],
                reg_name=register_name,
                reg_val_str=reg_val_str,
                optional_service=optional_service,
                optional_register=optional_register,
                timeout=timeout,
            )
    # The node is no longer needed.
    for msg in result.errors:
        show_error(msg)
    for msg in result.warnings:
        show_warning(msg)

    final = (
        {k: representer(v) for k, v in result.value_per_node.items()}
        if not isinstance(node_ids, int)
        else representer(result.value_per_node[node_ids])
    )
    sys.stdout.write(formatter(final))
    sys.stdout.flush()

    return EXIT_CODE_UNSUCCESSFUL if result.errors else 0


def _make_representer(simplify: bool, metadata: bool) -> Callable[[Optional["Access_1.Response"]], Any]:
    def represent(response: Optional["Access_1.Response"]) -> Any:
        return (
            explode_value(
                response.value,
                simplify=simplify,
                metadata=get_access_response_metadata(response) if metadata else None,
            )
            if response is not None
            else None
        )

    return represent



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

"""
The YAML library we use is API-unstable at the time of writing. We can't just use the de-facto standard PyYAML
because it's kinda stuck in the past (no ordered dicts, no support for YAML v1.2). This facade shields the
rest of the code from breaking changes in the YAML library API or from migration to another library.
"""

from ._dumper import Dumper as Dumper
from ._loader import Loader as Loader

from ._eval_loader import EvaluableLoader as EvaluableLoader
from ._eval_loader import EmbeddedExpressionError as EmbeddedExpressionError



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
from typing import Any, Callable, TypeVar
import decimal
import functools
import pycyphal


T = TypeVar("T")

METADATA_KEY = "_meta_"

EXIT_CODE_UNSUCCESSFUL = 100
"""
The command was invoked and executed correctly but the desired goal could not be attained for external reasons.
"""


def compose(*fs: Callable[..., T]) -> Callable[..., T]:
    """
    >>> compose(lambda x: x+2, lambda x: x*2)(3)
    10
    """
    return functools.reduce(_compose_unit, fs[::-1])


def _compose_unit(f: Callable[..., T], g: Callable[..., T]) -> Callable[..., T]:
    return lambda *a, **kw: f(g(*a, **kw))


def convert_transfer_metadata_to_builtin(
    transfer: pycyphal.transport.TransferFrom,
    *,
    dtype: Any,
    **extra_fields: dict[str, Any],
) -> dict[str, dict[str, Any]]:
    return {
        METADATA_KEY: {
            "ts_system": transfer.timestamp.system.quantize(_MICRO),
            "ts_monotonic": transfer.timestamp.monotonic.quantize(_MICRO),
            "source_node_id": transfer.source_node_id,
            "transfer_id": transfer.transfer_id,
            "priority": transfer.priority.name.lower(),
            "dtype": get_dtype_full_name_with_version(dtype),
            **extra_fields,
        }
    }


@functools.lru_cache(None)
def get_dtype_full_name_with_version(dtype: Any) -> str:
    return str(pycyphal.dsdl.get_model(dtype))


_MICRO = decimal.Decimal("0.000001")



# Copyright (c) 2019 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import re
from typing import Any, Type
import logging
import importlib

_logger = logging.getLogger(__name__)


class LoadError(Exception):
    pass


class FormatError(LoadError):
    pass


class NotFoundError(LoadError):
    pass


def load_dtype(name: str, allow_minor_version_mismatch: bool = False) -> Type[Any]:
    r"""
    Parses a data specifier string of the form ``full_data_type_name[.major_version[.minor_version]]``.
    Name separators may be replaced with ``/`` or ``\`` for compatibility with file system paths.
    Missing version numbers substituted with the latest one available.
    The short data type name is case-insensitive.

    :param name: Examples: ``uavcan.heartbeat``, ``uavcan.Heartbeat.1``, ``uavcan.HEARTBEAT.1.0``.

    :param allow_minor_version_mismatch:
        If the minor version is specified and there is no matching data type,
        repeat the search with the minor version omitted.
    """
    parsed = _parse(name)
    if not parsed:
        raise FormatError(f"Data type name format not understood: {name!r}")
    name_components, major, minor = parsed
    _logger.debug("Parsed %r: name_components=%r, major=%r, minor=%r", name, name_components, major, minor)
    try:
        result = _load(name_components, major, minor)
    except NotFoundError:
        if allow_minor_version_mismatch and minor is not None:
            result = _load(name_components, major, None)
        else:
            raise
    _logger.debug("Loaded %r as %r", name, result)
    return result


def _load(name_components: list[str], major: int | None, minor: int | None) -> Type[Any]:
    from yakut.cmd.compile import make_usage_suggestion

    namespaces, short_name = name_components[:-1], name_components[-1]
    try:
        mod = None
        for comp in namespaces:
            name = (mod.__name__ + "." + comp) if mod else comp  # type: ignore
            try:
                mod = importlib.import_module(name)
            except ImportError:  # We seem to have hit a reserved word; try with an underscore.
                mod = importlib.import_module(name + "_")
    except ImportError as ex:
        raise NotFoundError(make_usage_suggestion(namespaces[0])) from ex
    assert mod
    matches = sorted(
        (
            (x.groups(), getattr(mod, x.string))
            for x in filter(None, map(_RE_SHORT_TYPE_NAME_IDENTIFIER.match, dir(mod)))
            if (
                short_name.lower() == x.group(1).lower()
                and (major is None or int(x.group(2)) == major)
                and (minor is None or int(x.group(3)) == minor)
            )
        ),
        reverse=True,
    )
    _logger.debug("Identifiers in %r matching %s.%s.%s: %r", mod, short_name, major, minor, matches)
    if not matches:
        raise NotFoundError(
            f"Could not locate "
            f"{short_name}.{major if major is not None else '*'}.{minor if minor is not None else '*'} "
            f"in module {mod.__name__!r}"
        )
    return matches[0][1]  # type: ignore


def _parse(name: str) -> tuple[list[str], int | None, int | None] | None:
    m = _RE_PARSE.match(name)
    if not m:
        return None
    full_name, major, minor = m.groups()
    return list(_RE_SPLIT_NAME_COMPONENTS.split(full_name)), _version_or_nothing(major), _version_or_nothing(minor)


def _version_or_nothing(inp: str) -> int | None:
    return None if inp is None else int(inp)


_RE_PARSE = re.compile(
    r"^"
    r"([a-zA-Z_][a-zA-Z0-9_]*(?:[./\\][a-zA-Z_][a-zA-Z0-9_]*)+)"
    r"(?:[./\\](\d+))?"  # major version
    r"(?:[./\\](\d+))?"  # minor version
    r"$",
)
_RE_SPLIT_NAME_COMPONENTS = re.compile(r"\W")
_RE_SHORT_TYPE_NAME_IDENTIFIER = re.compile(r"([a-zA-Z_][a-zA-Z0-9_]*)_(\d+)_(\d+)")


def _unittest_re() -> None:
    m = _RE_PARSE.match("uavcan.node.Heartbeat.1.0")
    assert m and ("uavcan.node.Heartbeat", "1", "0") == m.groups()
    m = _RE_PARSE.match("uavcan/node\\Heartbeat/1\\0")
    assert m and ("uavcan/node\\Heartbeat", "1", "0") == m.groups()
    m = _RE_PARSE.match("uavcan.node.Heartbeat.1")
    assert m and ("uavcan.node.Heartbeat", "1", None) == m.groups()
    m = _RE_PARSE.match("uavcan.node.Heartbeat")
    assert m and ("uavcan.node.Heartbeat", None, None) == m.groups()
    m = _RE_PARSE.match("uavcan.Heartbeat")
    assert m and ("uavcan.Heartbeat", None, None) == m.groups()

    assert not _RE_PARSE.match("uavcan.node.Heartbeat.1.1.0")
    assert not _RE_PARSE.match("uavcan")


def _unittest_parse() -> None:
    assert (["uavcan", "node", "Heartbeat"], 1, 0) == _parse("uavcan.node.Heartbeat.1.0")
    assert (["uavcan", "node", "Heartbeat"], 1, None) == _parse("uavcan.node.Heartbeat.1")
    assert (["uavcan", "node", "Heartbeat"], None, None) == _parse("uavcan.node.Heartbeat")
    assert (["uavcan", "Heartbeat"], None, None) == _parse("uavcan.Heartbeat")



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
from typing import TYPE_CHECKING
import click
import pycyphal
import yakut
from yakut.int_set_parser import parse_int_set, INT_SET_USER_DOC
from yakut.ui import ProgressReporter, show_error, show_warning
from yakut.param.formatter import FormatterHints
from yakut.util import EXIT_CODE_UNSUCCESSFUL
from ._logic import list_names

if TYPE_CHECKING:
    import pycyphal.application

_logger = yakut.get_logger(__name__)


_HELP = f"""
List registers available on the specified remote node(s).

If the specified node-ID is a single integer, then the output contains simply the list of register names on that node.

If the specified node-ID is a set of integers (or a single integer followed by element separator, like `123,`),
then the output is a mapping (node_id->[register_name]).

The output can be piped to another command like yakut register-batch.

Examples:

\b
    y rl 42 > registers.json
    y rl 90,100..125,!110-115 | y rb > registers_all.json
    y rl '[1,2,42,105]' | jq > register_names_all.json

Filter by name, in this case those matching "uavcan*id" (mind the comma!):

\b
    y rl 125  | jq 'map(select(test("uavcan.+id")))'
    y rl 125, | jq 'map_values([.[] | select(test("uavcan.+id"))])'

Compute intersection -- registers that are available in all of the queried nodes:

\b
    y rl 120-128 | jq '. as $in|reduce .[] as $item ($in|flatten|flatten;.-(.-$item))|unique'

{INT_SET_USER_DOC}
"""


@yakut.subcommand(aliases="rl", help=_HELP)
@click.argument("node_ids", type=parse_int_set)
@click.option(
    "--timeout",
    "-T",
    type=float,
    default=pycyphal.presentation.DEFAULT_SERVICE_REQUEST_TIMEOUT,
    show_default=True,
    metavar="SECONDS",
    help="Service response timeout.",
)
@click.option(
    "--optional-service",
    "-s",
    is_flag=True,
    help="""
Ignore nodes that fail to respond to the first RPC-service request instead of reporting an error
assuming that the register service is not supported.
If a node responded at least once it is assumed to support the service and any future timeout
will be always treated as an error.
""",
)
@yakut.pass_purser
@yakut.asynchronous()
async def register_list(
    purser: yakut.Purser,
    node_ids: set[int] | int,
    timeout: float,
    optional_service: bool,
) -> int:
    _logger.debug("node_ids=%r, timeout=%r optional_service=%r", node_ids, timeout, optional_service)
    node_ids_list = list(sorted(node_ids)) if isinstance(node_ids, set) else [node_ids]
    assert isinstance(node_ids_list, list) and all(isinstance(x, int) for x in node_ids_list)
    formatter = purser.make_formatter(FormatterHints(single_document=True))
    with purser.get_node("register_list", allow_anonymous=False) as node:
        with ProgressReporter() as prog:
            result = await list_names(
                node,
                prog,
                node_ids_list,
                optional_service=optional_service,
                timeout=timeout,
            )
    # The node is no longer needed.
    for msg in result.errors:
        show_error(msg)
    for msg in result.warnings:
        show_warning(msg)
    final = result.names_per_node if not isinstance(node_ids, int) else result.names_per_node[node_ids]
    sys.stdout.write(formatter(final))
    sys.stdout.flush()

    return EXIT_CODE_UNSUCCESSFUL if result.errors else 0



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>


from __future__ import annotations
from typing import Any, Iterable
from pycyphal.presentation import Subscriber
from pycyphal.presentation.subscription_synchronizer.transfer_id import TransferIDSynchronizer
from ._sync import SynchronizerOutput, Synchronizer


def make_sync_transfer_id(subscribers: Iterable[Subscriber[Any]]) -> Synchronizer:
    sync = TransferIDSynchronizer(subscribers)

    async def fun(output: SynchronizerOutput) -> None:
        # noinspection PyTypeChecker
        async for synchronized_group in sync:
            output(synchronized_group)

    return fun



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import sys
import signal
import logging
from typing import Any
import click
import yakut

from ._schema import Composition as Composition
from ._schema import load_ast as load_ast
from ._schema import load_composition as load_composition

from ._executor import Context as Context
from ._executor import Stack as Stack
from ._executor import ErrorCode as ErrorCode
from ._executor import exec_file as exec_file
from ._executor import exec_composition as exec_composition


_logger = logging.getLogger(__name__)


def _indent(text: str) -> str:
    import textwrap

    return textwrap.indent(text, " " * 4)


# language=YAML
EXAMPLE_BASIC = """
$=:
- sleep 10
- echo $GREETING    # Concurrent execution!
-                   # Join statement, wait for "sleep" to finish.
- $=: echo $GREETING
  GREETING: bar     # Overrides the outer scope.
- ?=:               # This statement fails but error suppressed.
    $=: sleep 1
    .=: unknown-command-failure-ignored
-                   # Another join barrier.
- exit 88
.=: echo finalizer
GREETING: Hello world!
""".strip()
EXAMPLE_BASIC_STDOUT = """
Hello world!
bar
finalizer
""".strip()
EXAMPLE_BASIC_EXIT_CODE = 88


# language=YAML
EXAMPLE_EXTERNAL = """
external=:
- vars.orc.yaml   # Definition below
- echo.orc.yaml   # Definition below
.=: exit $EXIT_CODE
""".strip()
# language=YAML
EXAMPLE_EXTERNAL_VARS = """
# vars.orc.yaml
VARIABLE: This is my variable.
EXIT_CODE: 99
""".strip()
# language=YAML
EXAMPLE_EXTERNAL_ECHO = """
# echo.orc.yaml
$=: |
  python -c "from os import getenv; print(getenv('VARIABLE'))"
""".strip()
EXAMPLE_EXTERNAL_STDOUT = """
This is my variable.
""".strip()
EXAMPLE_EXTERNAL_EXIT_CODE = 99


# language=YAML
EXAMPLE_PUB_SUB = """
#!/usr/bin/env -S yakut orchestrate
# Compile DSDL and launch a pair of pub/sub.
$=:
- yakut compile $DSDL_SRC   # "DSDL_SRC" is to be set externally.
-                           # Wait for the compiler to finish.
- yakut --format json sub -M -N2 33:uavcan.si.unit.angle.Scalar.1.0
- yakut --format json sub -M -N1 uavcan.diagnostic.Record.1.1
- $=: >         # Inner composition with a single multi-line command.
    yakut pub -N5
    33:uavcan.si.unit.angle.Scalar.1.0 'radian: 4.0'
    uavcan.diagnostic.Record.1.1       'text: "four radians"'
  uavcan.node.id: 9  # The publisher is not anonymous unlike others.
.=:
- ?=: rm -r $YAKUT_COMPILE_OUTPUT       # Clean up at the exit.
uavcan.udp.iface: 127.42.0.1  # Configure the transport via registers.
YAKUT_COMPILE_OUTPUT: pub_sub_compiled_dsdl
YAKUT_PATH:           pub_sub_compiled_dsdl
""".strip()
# language=YAML
EXAMPLE_PUB_SUB_STDOUT = """
{"33":{"radian":4.0}}
{"8184":{"timestamp":{"microsecond":0},"severity":{"value":0},"text":"four radians"}}
{"33":{"radian":4.0}}
""".strip()


_HELP = f"""
Manage configuration of multiple Cyphal nodes using YAML files.

Currently, this tool is not tested against Windows and is therefore not expected to function there correctly.

Here, "orchestration" means configuration management of a Cyphal-based distributed computing system.
The participants of such a system may be either software processes executed on a computer (local or remote),
dedicated hardware units (e.g., a flight management unit or a sensor), or a mix thereof.

The user writes an orchestration file (orc-file for short) that defines the desired node parameters
and the shell commands that need to be executed to bring the system configuration into the desired state.
The configuration of remote nodes may be enforced via Cyphal using proxy commands like
"yakut register", "yakut call", etc. (this is convenient for hardware nodes) or via conventional means like SSH
(useful for software nodes executed on a remote computer).

As prescribed by the Cyphal standard, the node parameters are modeled as registers
(see standard RPC-service uavcan.register.Access).
Register values described in the orc-file are passed to the invoked processes via environment variables.
An environment variable name is constructed from register name by upper-casing it and replacing full stop characters
(".") with double low line characters ("__").

Binary-typed values (aka raw bytes) are passed as-is, strings are UTF8-encoded,
and numerical values are passed as decimals. Array elements are space-separated.
For example, register "m.motor.inductance_dq" of type "real32[2]" with value (0.12, 0.13)
is passed as an environment variable named "M__MOTOR__INDUCTANCE_DQ" assigned "0.12 0.13".

When defining a register value in the orc-file, the user may spell out its name in its entirety:

\b
    m.motor.inductance_dq: [0.12, 0.13]
    m.motor.flux_linkage:  1.34
    uavcan.node.id:        1201
    uavcan.pub.foo.id:     7777

For convenience and clarity, it is possible to group registers into dictionaries:

\b
    m.motor:
      inductance_dq: [0.12, 0.13]
      flux_linkage:   1.34
    uavcan:
      node.id: 1201
      pub.foo.id: 7777

It is also possible to define regular environment variables alongside the registers -- they will be passed
to the invoked processes as-is without modification.
Regular environment variables are distinguished from registers by the lack of "." in their names.

Null-valued keys can be used to erase registers and environment variables defined in an outer scope
(so that they are not propagated).

The behaviors defined by an orchestration file are specified using "compositions".
A composition is a YAML-dictionary of Cyphal registers, environment variables, and "directives".
A directive is a dictionary key ending with an equals sign "=".
The syntax can be approximated as follows, using a PEG-like notation:

\b
    # register and envvar are described in free form above.
    composition = (register / envvar / directive)+
    directive   = predicate / main / finalizer / external
    predicate   = "?=" : script
    main        = "$=" : script
    finalizer   = ".=" : script
    external    = "external=": files
    script      = statement / (statement*)
    statement   = shell / join / composition
    shell       = string
    join        = null
    files       = string / (string*)

The script directives "?=" (predicate), "$=" (main), and ".=" (finalizer) define shell commands or nested compositions
that are to be executed.
If there is more than one command/composition specified, all of them are launched concurrently
until the executor encounters the first "join statement", whereat it will wait for the previously
launched script items to complete successfully before continuing.

A shell command that returns zero is considered successful, otherwise it is considered to have failed.
The exit code of a composition is that of its first statement to fail, or zero if all completed successfully
(but see below about predicates).
The execution order is as follows:

1. The predicate script "?=" is executed first.
If any of its statements fail, its execution is aborted, pending processes are interrupted,
and success (sic!) is reported (that is, zero).
The predicate script is intended to define commands that are allowed/expected to fail,
hence the result is always zero.

2. The main script "$=" is executed only if none of the statements of the predicate have failed.
The first statement to fail aborts the execution of all pending statements and its exit code becomes
the exit code of the current composition.

3. The finalizer script ".=" is always executed if the predicate was successful after the main script,
even if the main script has failed.
Its exit code is discarded unless the main script executed successfully.
Execution of the finalizer script is never interrupted to ensure that the configuration of the managed
system is always left in a known state.
It is therefore recommended to avoid placing any non-trivial statements in the finalizer script.

Example:

\b
{_indent(EXAMPLE_BASIC)}

The exit code of the above composition is {EXAMPLE_BASIC_EXIT_CODE} and the stdout is:

\b
{_indent(EXAMPLE_BASIC_STDOUT)}

The "external=" directive defines an external orc-file to be executed BEFORE the script directives of the current
composition.
The path is either absolute or relative; in the latter case, the file will be searched for
in the current working directory and then through the directories specified in YAKUT_PATH.
The composition defined in the external file (the callee) inherits/overrides the environment variables
from the caller.
Upon successful execution, the caller inherits all environment variables back from the callee.
If the callee fails or cannot be executed for other reasons, the execution of the caller is aborted
and its exit code is propagated.
Example of three orc-files:

\b
{_indent(EXAMPLE_EXTERNAL)}

\b
{_indent(EXAMPLE_EXTERNAL_VARS)}

\b
{_indent(EXAMPLE_EXTERNAL_ECHO)}

The above example exits with code {EXAMPLE_EXTERNAL_EXIT_CODE} printing this:

\b
{_indent(EXAMPLE_EXTERNAL_STDOUT)}

The orchestrator uses the following exit codes to report internal errors:
125 -- orc-file is malformed;
124 -- orc-file not found or cannot be read.

The orchestrator attempts to manage the entire process tree atomically.
If execution of a process fails, the orchestrator will bring down its siblings, execute the finalizer,
and then return the exit code of the first failed process.
The orchestrator treats SIGINT/SIGTERM/SIGHUP sent to itself as a command to cease execution and finalize
the composition immediately.
When terminating a launched process, the orchestrator starts with SIGINT and gradually escalates up to SIGKILL
if the process refuses to stop peacefully in a reasonable time.

Children's stdout/stderr are piped into those of the orchestrator host process with per-line buffering
(so that concurrent output does not mangle lines of text).
Despite the line buffering, the streams are operated in binary mode.

A slightly more practical example is shown below:

\b
{_indent(EXAMPLE_PUB_SUB)}

The stdout of the above composition is:

\b
{_indent(EXAMPLE_PUB_SUB_STDOUT)}
"""


@yakut.subcommand(help=_HELP, aliases="orc")
@click.argument("file", type=str)
@yakut.pass_purser
def orchestrate(purser: yakut.Purser, file: str) -> int:
    sig_num = 0

    def on_signal(s: int, _: Any) -> None:
        nonlocal sig_num
        sig_num = s
        _logger.info("Orchestrator received signal %s %r, stopping...", s, signal.strsignal(s))

    signal.signal(signal.SIGINT, on_signal)
    signal.signal(signal.SIGTERM, on_signal)
    if not sys.platform.startswith("win"):
        signal.signal(signal.SIGHUP, on_signal)

    ctx = Context(lookup_paths=purser.paths)
    res = exec_file(ctx, file, {}, gate=lambda: sig_num == 0)

    return res if res != 0 else -sig_num



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import asyncio
from typing import Any, Iterable, Callable
import pycyphal
from pycyphal.transport import TransferFrom
from pycyphal.presentation import Subscriber
from ._sync import SynchronizerOutput, Synchronizer


def make_sync_async(subscribers: Iterable[Subscriber[Any]]) -> Synchronizer:
    """
    This synchronizer delivers one message at a time immediately without any synchronization,
    where all other messages and their metadata are set to None.
    """
    subscribers = list(subscribers)
    queue: asyncio.Queue[tuple[tuple[tuple[Any, TransferFrom] | None, Subscriber[Any]], ...]] = asyncio.Queue()

    def mk_handler(index: int) -> Callable[[Any, TransferFrom], None]:
        def hdl(msg: Any, meta: TransferFrom) -> None:
            assert isinstance(meta, TransferFrom)
            queue.put_nowait(
                tuple(
                    (
                        ((msg, meta) if idx == index else None),
                        sub,
                    )
                    for idx, sub in enumerate(subscribers)
                )
            )

        return hdl

    async def fun(output: SynchronizerOutput) -> None:
        try:
            for idx, sub in enumerate(subscribers):
                sub.receive_in_background(mk_handler(idx))
            while True:
                item = await queue.get()
                output(item)
        finally:
            pycyphal.util.broadcast((x.close for x in subscribers))()

    return fun


def _unittest_sync_async() -> None:
    from tests.dsdl import ensure_compiled_dsdl

    ensure_compiled_dsdl()

    from pycyphal.transport.loopback import LoopbackTransport
    from pycyphal.presentation import Presentation
    from uavcan.primitive.scalar import Integer8_1

    async def run() -> None:
        pre = Presentation(LoopbackTransport(10))
        sub_a = pre.make_subscriber(Integer8_1, 1000)
        sub_b = pre.make_subscriber(Integer8_1, 1001)
        pub_a = pre.make_publisher(sub_a.dtype, sub_a.port_id)
        pub_b = pre.make_publisher(sub_b.dtype, sub_b.port_id)
        try:
            syn = make_sync_async([sub_a, sub_b])
            results: list[tuple[tuple[tuple[Any, TransferFrom] | None, Subscriber[Any]], ...]] = []
            # noinspection PyTypeChecker
            tsk = asyncio.create_task(syn(results.append))
            try:
                await asyncio.sleep(0.1)
                assert not results

                await pub_a.publish(Integer8_1(50))
                await asyncio.sleep(0.1)
                ((((msg, meta), rx_sub_a), (none, rx_sub_b)),) = results  # type: ignore
                results.clear()
                assert rx_sub_a is sub_a and rx_sub_b is sub_b and none is None
                assert isinstance(msg, Integer8_1)
                assert msg.value == 50
                assert meta.source_node_id == 10

                await pub_a.publish(Integer8_1(51))
                await asyncio.sleep(0.1)
                ((((msg, meta), rx_sub_a), (none, rx_sub_b)),) = results  # type: ignore
                results.clear()
                assert rx_sub_a is sub_a and rx_sub_b is sub_b and none is None
                assert isinstance(msg, Integer8_1)
                assert msg.value == 51
                assert meta.source_node_id == 10

                await pub_b.publish(Integer8_1(52))
                await asyncio.sleep(0.1)
                (((none, rx_sub_a), ((msg, meta), rx_sub_b)),) = results  # type: ignore
                results.clear()
                assert rx_sub_a is sub_a and rx_sub_b is sub_b and none is None
                assert isinstance(msg, Integer8_1)
                assert msg.value == 52
                assert meta.source_node_id == 10

            finally:
                tsk.cancel()
        finally:
            sub_a.close()
            sub_b.close()
            pub_a.close()
            pub_b.close()
            pre.close()

    asyncio.run(run())



# Copyright (c) 2022 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>


from __future__ import annotations
from typing import Any, Callable, Awaitable, Iterable, Tuple, Optional
import pycyphal


SynchronizerOutput = Callable[
    [
        Tuple[
            Tuple[
                Optional[Tuple[Any, pycyphal.transport.TransferFrom]],
                pycyphal.presentation.Subscriber[Any],
            ],
            ...,
        ]
    ],
    None,
]

Synchronizer = Callable[
    [SynchronizerOutput],
    Awaitable[None],
]

SynchronizerFactory = Callable[
    [Iterable[pycyphal.presentation.Subscriber[Any]]],
    Synchronizer,
]



# Copyright (c) 2021 OpenCyphal
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@opencyphal.org>

from __future__ import annotations
import asyncio
import sys
from typing import Optional, Dict, TypeVar, Generic, cast, Any, TYPE_CHECKING
import shutil
import click
import pycyphal
from pycyphal.transport import MessageDataSpecifier, ServiceDataSpecifier, Timestamp, AlienTransfer
import yakut
from ._model import N_NODES, N_SUBJECTS, N_SERVICES, Avatar, NodeState
from ._view import View
from ._iface import Iface
from ._ui import refresh_screen

if TYPE_CHECKING:
    import pycyphal.application  # pylint: disable=ungrouped-imports


@yakut.subcommand(aliases="mon")
@click.option(
    "--plug-and-play",
    "-P",
    metavar="FILE",
    type=click.Path(dir_okay=False, resolve_path=True, path_type=str),
    help=f"""
Run a centralized plug-and-play (PnP) node-ID allocator alongside the monitor.
The file path points to the allocation table; if missing, a new file will be created.
If this option is used, the monitor cannot be used in anonymous mode (a local node-ID shall be given).
Low-level implementation details are available in the documentation for pycyphal.application.plug_and_play
at https://pycyphal.readthedocs.io.
""",
)
@yakut.pass_purser
@yakut.asynchronous(interrupted_ok=True)
async def monitor(purser: yakut.Purser, plug_and_play: Optional[str]) -> None:
    """
    Display information about online nodes and network traffic in real time (like htop).
    The command may attempt to detect and report some basic network configuration/implementation issues.

    If the tool is launched with an anonymous node-ID, its functionality will be limited
    as it will be unable to send introspection requests to other nodes (like uavcan.node.GetInfo),
    but it will always attempt to snoop on relevant data sent to other nodes and extract as much information as it can
    through passive monitoring (e.g., by intercepting responses to uavcan.node.GetInfo).

    The tool will set up low-level packet capture using the advanced network introspection API of PyCyphal.
    This may fail for some transports (UDP in particular) unless the user has the permissions required for packet
    capture and the system is configured appropriately (capabilities enabled, capture drivers installed, etc).
    Ethernet-based transports shall be connected to the SPAN port of the local switch (see port mirroring).
    Refer to the PyCyphal documentation for details: https://pycyphal.readthedocs.io.

    The output contains a table of nodes that are (or were) online followed by the connectivity matrix.
    The matrix shows the real-time traffic split by session:
    if node X emits transfers over port Y, the matrix cell (Y, X) will contain the averaged rate
    in transfers per second (t/s) for this session.
    The rightmost/bottom cells show the totals per port/node, respectively.
    These values allow one to assess how much load each element contributes to the total.
    The summary load in bytes per second (B/s) is also provided per port/node.

    Cells of the connectivity matrix are colored based on the port introspection information published
    by their node via uavcan.node.port.List.
    Cells that represent inactive ports (not a publisher, subscriber, client, or server) are colored black.
    Therefore, if a node publishes its port introspection and you see a value printed in a black cell,
    it is an error since the node is emitting transfers over a port that is not reported in the introspection
    message.

    Do keep in mind that the observed quantities are computed and valid for the application layer only,
    not the transport layer.
    Intentional or unintentional effects such as deterministic data loss mitigation, transport redundancy,
    or errors in the communication medium may duplicate transfers or transport frames (packets) exchanged
    over the network, but such occurrences will not be displayed on the matrix because it is by design
    supposed to focus on the application-layer workload.
    If you need a lower-level view of the network, consider using specialized tools such as Wireshark or candump.
    """
    # pylint: disable=too-many-locals
    import numpy as np
    from scipy.sparse import dok_matrix, spmatrix

    try:
        import uavcan.node.port  # pylint: disable=unused-import
    except ImportError as ex:
        from yakut.cmd.compile import make_usage_suggestion

        raise click.ClickException(make_usage_suggestion(ex.name))

    allow_anonymous = not plug_and_play

    period = 2.0

    fir_window_duration = 10.0
    fir_samples = max(2, round(fir_window_duration / period))
    xfer_rates_filter: MovingAverage[spmatrix] = MovingAverage(fir_samples)
    byte_rates_filter: MovingAverage[spmatrix] = MovingAverage(fir_samples)

    total_transport_error_count = 0

    models: Dict[Optional[int], Avatar] = {}

    ts_last_trace: Optional[Timestamp] = None

    # We have to keep all counters in a sparse matrix for performance reasons,
    # otherwise we observe bad performance issues even with small networks.
    # The anonymous node stat is located at the max index.
    xfer_counts = dok_matrix((N_NODES + 1, N_SUBJECTS + N_SERVICES * 2), dtype=np.uint64)
    xfer_counts_prev = xfer_counts.copy()
    byte_counts = dok_matrix((N_NODES + 1, N_SUBJECTS + N_SERVICES * 2), dtype=np.uint64)
    byte_counts_prev = byte_counts.copy()

    def on_trace(ts: Timestamp, tr: AlienTransfer) -> None:
        nonlocal ts_last_trace
        ts_last_trace = ts
        # Create node instances lazily.
        node_id = tr.metadata.session_specifier.source_node_id
        if node_id not in models:
            _logger.info("New node %r", node_id)
            models[node_id] = Avatar(iface, node_id)
        # Update statistical counters.
        x = node_id
        if x is None:
            x = N_NODES
        y = linearize_data_specifier(tr.metadata.session_specifier.data_specifier)
        xfer_counts[x, y] += 1
        byte_counts[x, y] += sum(map(len, tr.fragmented_payload))

    def on_transport_error(tr: pycyphal.transport.ErrorTrace) -> None:
        nonlocal total_transport_error_count
        total_transport_error_count += 1
        _ = tr

    with purser.get_node("monitor", allow_anonymous=allow_anonymous) as node:
        iface = Iface(node)
        iface.add_trace_handler(on_trace)
        iface.add_transport_error_handler(on_transport_error)

        if plug_and_play:
            _logger.info("Starting plug-and-play allocator using file %r", plug_and_play)
            launch_centralized_plug_and_play_allocator(plug_and_play, node)

        # Special case: set up an entry for the local node manually because we won't be able to request own node.
        if node.id is not None:
            models[node.id] = Avatar(iface, node.id, info=node.info)

        _logger.debug("STARTING THE MAIN LOOP")
        loop = asyncio.get_event_loop()
        view = View()
        next_redraw_at = loop.time()
        ts_prev = next_redraw_at
        while True:
            next_redraw_at += period
            await asyncio.sleep(next_redraw_at - loop.time())
            ts_started = loop.time()

            # If there were any network events since last update, use the latest time received from the network.
            # Only if there were no updates use the real time, which may happen IFF the local node is anonymous
            # and the network is completely silent (so not a typical scenario).
            # The point of this manipulation is to use the correct time when the tool has accumulated a long input
            # frame queue due to being unable to keep up with data processing in real time (hey, we're using Python).
            # In this case, using the real wall time would be harmful as it would lead the tool to believe that
            # the data is obsolete whereas it's the tool's view of time that is obsolete rather than the data itself.
            ts = float(ts_last_trace.monotonic) if ts_last_trace is not None else ts_started
            ts_last_trace = None
            dt_trace = ts - ts_prev
            ts_prev = ts

            # Update statistics
            xfer_deltas = xfer_counts - xfer_counts_prev
            xfer_rates_filter.update(xfer_deltas.astype(np.float64) / dt_trace)
            xfer_counts_prev = xfer_counts.copy()
            xfer_rates = xfer_rates_filter.compute()
            byte_rates_filter.update((byte_counts - byte_counts_prev).astype(np.float64) / dt_trace)
            byte_counts_prev = byte_counts.copy()
            byte_rates = byte_rates_filter.compute()
            elapsed_stats = loop.time() - ts_started

            # Recompute the new node state snapshots
            snapshot: Dict[Optional[int], NodeState] = {}
            for node_id, mo in sorted(models.items(), key=lambda x: x[0] if x[0] is not None else N_NODES):
                assert isinstance(mo, Avatar)
                snapshot[node_id] = mo.update(ts)

            # Render the model into a huge text buffer
            ts_render_started = loop.time()
            view.render(
                states=snapshot,
                xfer_deltas=xfer_deltas,
                xfer_rates=xfer_rates,
                byte_rates=byte_rates,
                total_transport_errors=total_transport_error_count,
                fir_window_duration=fir_window_duration,
                max_width_height=shutil.get_terminal_size((sys.maxsize, sys.maxsize)),
            )
            elapsed_render = loop.time() - ts_render_started

            # Send the text buffer to the terminal (may block)
            await loop.run_in_executor(None, refresh_screen, view.flip_buffer())
            _logger.info(
                "Updated in %.3f (stats %.3f, render %.3f); dt %.3f; time lag %.3f",
                loop.time() - ts_started,
                elapsed_stats,
                elapsed_render,
                dt_trace,
                ts_started - ts,
            )


def linearize_data_specifier(ds: pycyphal.transport.DataSpecifier) -> int:
    if isinstance(ds, MessageDataSpecifier):
        return int(ds.subject_id)
    if isinstance(ds, ServiceDataSpecifier):
        if ds.role == ServiceDataSpecifier.Role.REQUEST:
            return int(ds.service_id + N_SUBJECTS)
        if ds.role == ServiceDataSpecifier.Role.RESPONSE:
            return int(ds.service_id + N_SUBJECTS + N_SERVICES)
    assert False


_T = TypeVar("_T")


class MovingAverage(Generic[_T]):
    """
    Beware that this implementation trades off numerical stability for speed.
    This is adequate for a visualization tool but may be unsuitable in most other applications.
    The reason it is unstable is that instead of recomputing the sum at every iteration
    it keeps a running sum whose delta is updated at every update.
    The delta update may cause accumulation of numerical errors over time.
    """

    def __init__(self, depth: int) -> None:
        from collections import deque

        self._depth = int(depth)
        self._samples: deque[_T] = deque(maxlen=depth)
        self._sum: Any = None

    def update(self, x: _T) -> None:
        if x is None:
            raise ValueError(repr(x))
        if self._sum is not None:
            if len(self._samples) >= self._depth:
                self._sum -= self._samples.popleft()
            self._sum += x
        else:
            from copy import copy

            self._sum = copy(x)
        self._samples.append(x)

    def compute(self) -> _T:
        return cast(_T, self._sum * (1.0 / len(self._samples)))


def launch_centralized_plug_and_play_allocator(database_path: str, node: pycyphal.application.Node) -> None:
    from pycyphal.application.plug_and_play import CentralizedAllocator
    from pycyphal.application.node_tracker import NodeTracker, Entry

    alloc = CentralizedAllocator(node, database_path)
    tracker = NodeTracker(node)

    def register_node(node_id: int, _old_entry: Optional[Entry], entry: Optional[Entry]) -> None:
        if entry is not None:
            unique_id = entry.info.unique_id.tobytes() if entry.info else None
            alloc.register_node(node_id, unique_id)

    tracker.add_update_handler(register_node)


_logger = yakut.get_logger(__name__)
